from tests.utils import TEST_DATA_DIR
from dexy.doc import Doc
from tests.utils import wrap
import os

markdown_file = os.path.join(TEST_DATA_DIR, "markdown-test.md")

def run_kramdown(ext):
    with open(markdown_file, 'r') as f:
        example_markdown = f.read()

    with wrap() as wrapper:
        node = Doc("markdown.md|kramdown",
                wrapper,
                [],
                kramdown = { 'ext' : ext },
                contents = example_markdown
                )
        wrapper.run_docs(node)
        assert node.output_data().is_cached()
        return node.output_data()

def test_kramdown_html():
    html = str(run_kramdown(".html"))
    assert """<h2 id="download">""" in html

def test_kramdown_tex():
    tex = str(run_kramdown(".tex"))
    assert "\subsection" in tex



from tests.utils import tempdir
from nose.exc import SkipTest

try:
    from dexy_filter_examples import Cowsay
except ImportError:
    raise SkipTest()

raise SkipTest() # TODO fixme

def test_cowsay():
    with tempdir():
        for batch in Cowsay().dexy():
            print(batch)



from tests.utils import TEST_DATA_DIR
from dexy.doc import Doc
from tests.utils import wrap
import os

sweave_file = os.path.join(TEST_DATA_DIR, "example-2.Snw")

def test_sweave_filter():
    with open(sweave_file, 'r') as f:
        sweave_content = f.read()

    with wrap() as wrapper:
        node = Doc("example.Snw|sweave",
                wrapper,
                [],
                contents = sweave_content
              )
        wrapper.run_docs(node)
        assert node.output_data().is_cached()
        assert "Coefficients:" in str(node.output_data())



from tests.utils import runfilter
from tests.utils import wrap
from dexy.doc import Doc

XML = """
<element id="foo">foo</element>
"""

def test_xxml():
    with runfilter("xxml", XML) as doc:
        assert doc.output_data()['foo:source'] == '<element id="foo">foo</element>'
        assert doc.output_data()['foo:text'] == 'foo'
        assert '<div class="highlight">' in doc.output_data()['foo:html-source']

def test_xxml_no_pygments():
    with wrap() as wrapper:
        doc = Doc(
                "example.xml|xxml",
                wrapper,
                [],
                contents = XML,
                xxml = { 'pygments' : False, 'ext' : '.sqlite3' }
                )
        wrapper.run_docs(doc)

        assert "foo:source" in list(doc.output_data().keys())
        assert not "foo:html-source" in list(doc.output_data().keys())
        assert not "foo:latex-source" in list(doc.output_data().keys())



from dexy.filter import DexyFilter

try:
    import bleach
    AVAILABLE = True
except ImportError:
    AVAILABLE = False

class Bleach(DexyFilter):
    """
    Runs the Bleach HTML sanitizer. <https://github.com/jsocol/bleach>
    """
    aliases = ['bleach']

    _settings = {
            'added-in-version' : '0.9.9.6',
            'input-extensions' : ['.html', '.txt'],
            'output-extensions' : ['.html', '.txt']
            }

    def is_active(self):
        return AVAILABLE

    # TODO implement support for sections
    def process_text(self, input_text):
        return bleach.clean(input_text)



from dexy.commands.utils import init_wrapper
from dexy.exceptions import InternalDexyProblem
from dexy.exceptions import UserFeedback
from dexy.parser import AbstractSyntaxTree
from dexy.parsers.doc import Yaml
from tests.utils import capture_stdout
from tests.utils import tempdir
from tests.utils import wrap
from dexy.wrapper import Wrapper
import dexy.batch
import os

def test_deprecated_dot_dexy_file():
    with tempdir():
        with open(".dexy", 'w') as f:
            f.write("{}")
        wrapper = Wrapper()
        try:
            wrapper.assert_dexy_dirs_exist()
        except UserFeedback as e:
            assert "this format is no longer supported" in str(e)

def test_cache_and_dexy_dirs_present():
    with tempdir():
        os.mkdir(".dexy")
        os.mkdir(".cache")
        with open(".dexy/.dexy-generated", 'w') as f:
            f.write("")
        with open(".cache/.dexy-generated", 'w') as f:
            f.write("")

        wrapper = Wrapper()

        try:
            wrapper.assert_dexy_dirs_exist()
        except UserFeedback as e:
            assert "Please remove '.cache'" in str(e)

        os.remove(".cache/.dexy-generated")
        wrapper.assert_dexy_dirs_exist()

        # Cache still exists but dexy just ignores it.
        assert os.path.exists(".cache")

        # Dexy uses .dexy dir
        assert os.path.exists(".dexy")

def test_move_cache_dir():
    with capture_stdout() as stdout:
        with tempdir():
            os.mkdir(".cache")
            with open(".cache/.dexy-generated", 'w') as f:
                f.write("")
    
            wrapper = Wrapper()
            wrapper.assert_dexy_dirs_exist()
    
            assert "Moving directory '.cache'" in stdout.getvalue()
            assert not os.path.exists(".cache")
            assert os.path.exists(".dexy")

def test_old_cache_dir_with_settings():
    with capture_stdout() as stdout:
        with tempdir():
            os.mkdir(".cache")

            with open(".cache/.dexy-generated", 'w') as f:
                f.write("")
    
            wrapper = Wrapper(artifacts_dir = ".cache")
            wrapper.assert_dexy_dirs_exist()

            assert os.path.exists(".cache")
            assert not os.path.exists(".dexy")
    
            assert "You may have a dexy.conf file" in stdout.getvalue()

def test_remove_trash_no_trash():
    with tempdir():
        wrapper = Wrapper()
        wrapper.empty_trash()

def test_remove_trash_with_trash():
    with tempdir():
        wrapper = Wrapper()
        os.mkdir(".trash")
        assert os.path.exists(".trash")
        wrapper.empty_trash()
        assert not os.path.exists(".trash")

def test_state_new_after_init():
    wrapper = Wrapper()
    wrapper.validate_state('new')

def test_error_if_to_valid_called_without_dirs_setup():
    with tempdir():
        wrapper = Wrapper()
        try:
            wrapper.to_valid()
            assert False, "should not get here"
        except InternalDexyProblem:
            assert True

def test_state_valid_after_to_valid():
    with tempdir():
        wrapper = Wrapper()
        wrapper.create_dexy_dirs()
        wrapper.to_valid()
        wrapper.validate_state('valid')

def test_walked():
    with tempdir():
        with open("dexy.yaml", "w") as f:
            f.write("foo.txt")

        with open("foo.txt", "w") as f:
            f.write("foo")

        wrapper = Wrapper()
        wrapper.create_dexy_dirs()
        wrapper.to_valid()
        wrapper.to_walked()
        wrapper.validate_state('walked')

def test_checked():
    with tempdir():
        with open("dexy.yaml", "w") as f:
            f.write("foo.txt")

        with open("foo.txt", "w") as f:
            f.write("foo")

        wrapper = Wrapper()
        wrapper.create_dexy_dirs()
        wrapper.to_valid()
        wrapper.to_walked()
        wrapper.to_checked()
        wrapper.validate_state('checked')

def test_ran():
    with tempdir():
        with open("dexy.yaml", "w") as f:
            f.write("foo.txt")

        with open("foo.txt", "w") as f:
            f.write("foo")

        wrapper = Wrapper()
        wrapper.create_dexy_dirs()
        wrapper.run_from_new()
        for node in wrapper.roots:
            assert node.state == 'ran'
        wrapper.validate_state('ran')

        wrapper = Wrapper()
        wrapper.run_from_new()
        for node in wrapper.roots:
            assert node.state == 'consolidated'
        wrapper.validate_state('ran')

def test_explicit_configs():
    wrapper = Wrapper()
    wrapper.configs = "foo.txt bar.txt   abc/def/foo.txt "
    assert wrapper.explicit_config_files() == ['foo.txt',
            'bar.txt', 'abc/def/foo.txt']

def test_parse_doc_configs_single_empty_config():
    with tempdir():
        wrapper = Wrapper()
        wrapper.create_dexy_dirs()

        with open("dexy.yaml", "w") as f:
            f.write("foo.txt")

        with open("foo.txt", "w") as f:
            f.write("foo")

        wrapper = Wrapper()
        wrapper.to_valid()
        wrapper.to_walked()

def test_parse_doc_configs_no_configs():
    with tempdir():
        with capture_stdout() as stdout:
            wrapper = Wrapper()
            wrapper.create_dexy_dirs()

            wrapper = Wrapper()
            wrapper.to_valid()
            wrapper.to_walked()
            value = stdout.getvalue()
        assert "didn't find any document config files" in value

def test_assert_dexy_dirs():
    with tempdir():
        wrapper = Wrapper()
        try:
            wrapper.assert_dexy_dirs_exist()
            assert False
        except UserFeedback:
            assert True

def test_create_remove_dexy_dirs():
    with tempdir():
        wrapper = Wrapper()
        wrapper.create_dexy_dirs()
        wrapper.to_valid()
        assert wrapper.dexy_dirs_exist()
        wrapper.remove_dexy_dirs()
        assert not wrapper.dexy_dirs_exist()

def test_init_wrapper_if_dexy_dirs_exist():
    with tempdir():
        wrapper = Wrapper()
        wrapper.create_dexy_dirs()

        with open("hello.txt", "w") as f:
            f.write("hello")

        wrapper = Wrapper()
        wrapper.to_valid()
        assert wrapper.project_root
        wrapper.to_walked()
        assert 'hello.txt' in wrapper.filemap
        assert 'dexy.log' in os.listdir('.dexy')
        assert not '.dexy/dexy.log' in wrapper.filemap

def test_nodexy_files():
    with tempdir():
        wrapper = Wrapper()
        wrapper.create_dexy_dirs()

        with open("hello.txt", "w") as f:
            f.write("hello")

        os.makedirs("s1/s2/s3")

        nodexy_path = "s1/s2/.nodexy"
        with open(nodexy_path, 'w') as f:
            f.write("dexy stop here")

        with open("s1/s2/ignore.txt", "w") as f:
            f.write("dexy should ignore this")

        with open("s1/s2/s3/ignore.txt", "w") as f:
            f.write("dexy should also ignore this")

        # Only the hello.txt file is visible to dexy
        wrapper = Wrapper()
        wrapper.to_valid()
        wrapper.to_walked()
        assert len(wrapper.filemap) == 1
        assert 'hello.txt' in wrapper.filemap

        os.remove(nodexy_path)

        # Now we can see all 3 text files.
        wrapper = Wrapper()
        wrapper.to_valid()
        wrapper.to_walked()
        assert len(wrapper.filemap) == 3
        assert 'hello.txt' in wrapper.filemap
        assert 's1/s2/ignore.txt' in wrapper.filemap
        assert 's1/s2/s3/ignore.txt' in wrapper.filemap

# old
def test_config_for_directory():
    with wrap() as wrapper:
        with open("dexy.yaml", "w") as f:
            f.write(""".abc""")

        with open("root.abc", "w") as f:
            f.write("hello")

        with open("root.def", "w") as f:
            f.write("hello")

        os.makedirs("s1")
        os.makedirs("s2")

        with open("s1/s1.abc", "w") as f:
            f.write("hello")

        with open("s1/s1.def", "w") as f:
            f.write("hello")

        with open("s2/s2.abc", "w") as f:
            f.write("hello")

        with open("s2/s2.def", "w") as f:
            f.write("hello")

        with open(os.path.join('s1', 'dexy.yaml'), 'w') as f:
            f.write(""".def|dexy""")

        wrapper = Wrapper()
        wrapper.to_valid()
        wrapper.to_walked()
        wrapper.to_checked()
        wrapper.run()

        assert len(wrapper.nodes) == 6

        p = wrapper.nodes["pattern:*.abc"]
        c = wrapper.nodes["doc:s2/s2.abc"]
        assert c in p.children

def test_config_file():
    with tempdir():
        with open("dexy.conf", "w") as f:
            f.write("""{ "logfile" : "a.log" }""")

        wrapper = init_wrapper({'conf' : 'dexy.conf'})
        assert wrapper.log_file == "a.log"

def test_kwargs_override_config_file():
    with tempdir():
        with open("dexy.conf", "w") as f:
            f.write("""{ "logfile" : "a.log" }""")

        wrapper = init_wrapper({
            '__cli_options' : { 'logfile' : 'b.log' },
            'logfile' : "b.log",
            'conf' : 'dexy.conf'
            })
        assert wrapper.log_file == "b.log"

def test_wrapper_init():
    wrapper = Wrapper()
    assert wrapper.artifacts_dir == '.dexy'

YAML = """foo:
    - bar
    - baz

foob:
    - foobar

xyz:
    - abc
    - def
"""

def run_yaml_with_target(target):
    with wrap() as wrapper:
        wrapper.nodes = {}
        wrapper.roots = []
        wrapper.batch = dexy.batch.Batch(wrapper)
        wrapper.filemap = wrapper.map_files()

        ast = AbstractSyntaxTree(wrapper)
        parser = Yaml(wrapper, ast)
        parser.parse('.', YAML)
        ast.walk()

        wrapper.transition('walked')
        wrapper.to_checked()

        assert len(wrapper.roots) == 3
        assert len(wrapper.nodes) == 8

        wrapper.target = target
        wrapper.run()

        yield wrapper

def test_run_target_foo():
    for wrapper in run_yaml_with_target("foo"):
        assert wrapper.nodes['bundle:foo'].state == 'ran'
        assert wrapper.nodes['bundle:bar'].state == 'ran'
        assert wrapper.nodes['bundle:baz'].state == 'ran'
        assert wrapper.nodes['bundle:foob'].state == 'uncached'
        assert wrapper.nodes['bundle:foobar'].state == 'uncached'
        assert wrapper.nodes['bundle:xyz'].state == 'uncached'

def test_run_target_fo():
    for wrapper in run_yaml_with_target("fo"):
        # foo and children have been run
        assert wrapper.nodes['bundle:foo'].state == 'ran'
        assert wrapper.nodes['bundle:bar'].state == 'ran'
        assert wrapper.nodes['bundle:baz'].state == 'ran'

        # foob and children have been run
        assert wrapper.nodes['bundle:foob'].state == 'ran'
        assert wrapper.nodes['bundle:foobar'].state == 'ran'

def test_run_target_bar():
    for wrapper in run_yaml_with_target("bar"):
        assert wrapper.nodes['bundle:foo'].state == 'uncached'
        assert wrapper.nodes['bundle:bar'].state == 'ran'
        assert wrapper.nodes['bundle:baz'].state == 'uncached'
        assert wrapper.nodes['bundle:foob'].state == 'uncached'
        assert wrapper.nodes['bundle:foobar'].state == 'uncached'

def test_run_target_ba():
    for wrapper in run_yaml_with_target("ba"):
        assert wrapper.nodes['bundle:foo'].state == 'uncached'
        assert wrapper.nodes['bundle:bar'].state == 'ran'
        assert wrapper.nodes['bundle:baz'].state == 'ran'
        assert wrapper.nodes['bundle:foob'].state == 'uncached'
        assert wrapper.nodes['bundle:foobar'].state == 'uncached'



from dexy.templates.standard import DefaultTemplate
from tests.utils import tempdir
import os

def test_run():
    with tempdir():
        DefaultTemplate().generate("test")
        assert not os.path.exists("dexy.rst")

def test_dexy():
    for wrapper in DefaultTemplate().dexy():
        batch = wrapper.batch
        assert 'jinja' in batch.filters_used
        assert "doc:hello.txt|jinja" in batch.docs
        assert "doc:dexy.rst|jinja|rst2html" in batch.docs

def test_validate_default():
    DefaultTemplate().validate()



from dexy.doc import Doc
from tests.utils import wrap
from dexy.wrapper import Wrapper

def test_yamlargs_with_caching():
    with wrap() as wrapper:
        doc = Doc("example.txt|yamlargs",
                wrapper,
                [],
                contents = "title: My Title\n---\r\nThis is the content."
                )
        wrapper.run_docs(doc)

        task = wrapper.nodes["doc:example.txt|yamlargs"]
        assert task.output_data().title() == "My Title"
        assert task.state == 'ran'

        wrapper = Wrapper()
        doc = Doc("example.txt|yamlargs",
                wrapper,
                [],
                contents = "title: My Title\n---\r\nThis is the content."
                )
        wrapper.run_docs(doc)
        task = wrapper.nodes["doc:example.txt|yamlargs"]
        assert task.output_data().title() == "My Title"
        assert task.state == 'consolidated'

        wrapper = Wrapper()
        doc = Doc("example.txt|yamlargs",
                wrapper,
                [],
                contents = "title: My Title\n---\r\nThis is the content."
                )
        wrapper.run_docs(doc)
        task = wrapper.nodes["doc:example.txt|yamlargs"]
        assert task.output_data().title() == "My Title"
        assert task.state == 'consolidated'

def test_yamlargs_no_yaml():
    with wrap() as wrapper:
        doc = Doc("example.txt|yamlargs",
                wrapper,
                [],
                contents = "This is the content.")

        wrapper.run_docs(doc)
        assert doc.output_data().as_text() == "This is the content."

def test_yamlargs():
    with wrap() as wrapper:
        doc = Doc("example.txt|yamlargs",
                wrapper,
                [],
                contents = "title: My Title\n---\r\nThis is the content."
                )

        wrapper.run_docs(doc)
        assert doc.output_data().title() == "My Title"
        assert doc.output_data().as_text() == "This is the content."

YAML = """filterargs:
  abc: xyz
  foo: 5
"""

def test_yamlargs_filterargs():
    with wrap() as wrapper:
        doc = Doc("example.txt|yamlargs|filterargs",
                wrapper,
                [],
                contents = "%s\n---\r\nThis is the content." % YAML,
                )

        wrapper.run_docs(doc)

        output = doc.output_data().as_text()
        assert "abc: xyz" in output
        assert "foo: 5" in output

        wrapper = Wrapper()
        doc = Doc("example.txt|yamlargs|filterargs",
                wrapper,
                [],
                contents = "%s\n---\r\nThis is the content." % YAML,
                )

        wrapper.run_docs(doc)

        output = doc.output_data().as_text()
        assert "abc: xyz" in output
        assert "foo: 5" in output



from tests.utils import assert_output
from tests.utils import assert_in_output
from tests.utils import wrap
from dexy.doc import Doc

rst_meta = """
==========
Main Title
==========

---
Foo
---

:Author: J Random Hacker
:Authors: Bert & Ernie
:Contact: jrh@example.com
:Date: 2002-08-18
:Status: Work In Progress
:Version: 1
:Filename: $RCSfile$
:Copyright: This document has been placed in the public domain.

Here's some content.

"""
def test_rst_meta():
    with wrap() as wrapper:
        node = Doc("example.rst|rstmeta",
                wrapper, 
                [],
                contents = rst_meta
                )
        wrapper.run_docs(node)

        assert node.setting('author') == "J Random Hacker"
        assert node.setting('authors') == "Bert & Ernie"
        assert node.setting('subtitle') == "Foo"
        assert node.setting('title') == "Main Title"
        assert node.setting('date') == "2002-08-18"
        assert node.setting('status') == "Work In Progress"
        assert node.setting('version') == "1"
        assert node.setting('copyright') == "This document has been placed in the public domain."

RST = """
* a bullet point using "*"

  - a sub-list using "-"

    + yet another sub-list

  - another item
"""

def test_rst2odt():
    with wrap() as wrapper:
        node = Doc("example.txt|rst2odt",
                wrapper,
                [],
                contents=RST)
        wrapper.run_docs(node)
        assert node.output_data().filesize() > 8000

def test_rst2xml():
    assert_in_output('rst2xml', RST, """<list_item><paragraph>a sub-list using "-"</paragraph><bullet_list bullet="+"><list_item>""")

def test_rst2latex():
    assert_in_output('rst2latex', RST, "\item a bullet point using")
    assert_in_output('rst2latex', RST, "\\begin{document}")

def test_rst2html():
    assert_in_output('rst2html', RST, "<html xmlns")
    assert_in_output('rst2html', RST, "<li>a bullet point using &quot;*&quot;<ul>")

def test_rest_to_tex():
    with wrap() as wrapper:
        node = Doc("example.txt|rstbody",
                wrapper,
                [],
                contents=RST,
                rstbody={"ext" : ".tex"}
                )

        wrapper.run_docs(node)
        assert "\\begin{itemize}" in str(node.output_data())

def test_rest_to_html():
    expected = """\
<ul class="simple">
<li>a bullet point using &quot;*&quot;<ul>
<li>a sub-list using &quot;-&quot;<ul>
<li>yet another sub-list</li>
</ul>
</li>
<li>another item</li>
</ul>
</li>
</ul>
"""

    assert_output('rstbody', RST, expected)

def test_rstbody_latex():
    with wrap() as wrapper:
        node = Doc("example.rst|rstbody",
                wrapper, 
                [],
                rstbody = { 'ext' : '.tex' },
                contents = RST
                )
        wrapper.run_docs(node)
        output = str(node.output_data())
        assert "\\begin{itemize}" in output



from dexy.filter import DexyFilter
import os
import re

class SplitHtmlFilter(DexyFilter):
    """
    Generate index page linking to multiple pages from single source.

    The split filter looks for specially formatted HTML comments in your
    document and splits your HTML into separate pages at each split comment.
    """
    aliases = ['split', 'splithtml']
    _settings = {
            'output' : True,
            'input-extensions' : ['.html'],
            'output-extensions' : ['.html'],
            'split-ul-class' : ("HTML class to apply to <ul> elements", None)
            }

    def process(self):
        input_text = str(self.input_data)

        if input_text.find("<!-- endsplit -->") > 0:
            rawbody, footer = re.split("<!-- endsplit -->", input_text, maxsplit=1)
           
            if rawbody.find("<!-- footer -->") > 0:
                body, index_footer_content = re.split("<!-- footer -->", rawbody, maxsplit=1) 
            else:
                body = rawbody
                index_footer_content = ""

            sections = re.split("<!-- split \"(.+)\" -->\n", body)
            header = sections[0]

            pages = {}
            index_top_content = ""
            for i in range(1, len(sections), 2):
                if sections[i] == 'index':
                    index_top_content = sections[i+1]
                else:
                    section_label = sections[i]
                    section_content = sections[i+1]

                    section_url = section_label.split(" ")[0]

                    if "<!-- content -->" in section_content:
                        section_description, section_content = section_content.split("<!-- content -->")
                    else:
                        section_description = ""

                    filename = "%s.html" % section_url

                    filepath = os.path.join(self.output_data.parent_dir(), filename)
                    pages[section_label] = (section_description, filename,)

                    new_page = self.add_doc(filepath, header + section_content + footer)
                    new_page.update_setting('title', re.sub("\s+\(.+\)\s*(<.*>\s*)*", "", section_label))

                    apply_ws_to_content = self.doc.safe_setting("apply-ws-to-content")
                    apply_ws_to_content_variable_start_string = self.doc.safe_setting("apply-ws-to-content-variable-start-string")
                    apply_ws_to_content_variable_end_string = self.doc.safe_setting("apply-ws-to-content-variable-end-string")
                    apply_ws_to_content_block_start_string = self.doc.safe_setting("apply-ws-to-content-block-start-string")
                    apply_ws_to_content_block_end_string = self.doc.safe_setting("apply-ws-to-content-block-end-string")

                    if apply_ws_to_content:
                        new_page.update_setting('apply-ws-to-content', apply_ws_to_content)
                    if apply_ws_to_content_variable_start_string:
                        new_page.update_setting('apply-ws-to-content-variable-start-string', apply_ws_to_content_variable_start_string)
                    if apply_ws_to_content_variable_end_string:
                        new_page.update_setting('apply-ws-to-content-variable-end-string', apply_ws_to_content_variable_end_string)
                    if apply_ws_to_content_block_start_string:
                        new_page.update_setting('apply-ws-to-content-block-start-string', apply_ws_to_content_block_start_string)
                    if apply_ws_to_content_block_end_string:
                        new_page.update_setting('apply-ws-to-content-block-end-string', apply_ws_to_content_block_end_string)

                    self.log_debug("added key %s to %s ; links to file %s" %
                              (filepath, self.key, new_page.name))

            index_items = ""
            for page_label in sorted(pages):
                page_description, filename = pages[page_label]
                index_items += """<li><a href="%s">%s</a></li>\n%s\n""" % (filename, page_label, page_description)

            output = header + index_top_content

            if self.setting("split-ul-class"):
                ul = "<ul class=\"%s\">" % self.setting('split-ul-class')
            else:
                ul = "<ul class=\"split\">"

            output += "%s\n%s\n</ul>" % (ul, index_items)
            output += index_footer_content + footer

        else:
            # No endsplit found, do nothing.
            output = input_text

        self.output_data.set_data(output)



from tests.utils import runfilter
from tests.utils import wrap
from dexy.doc import Doc
from nose.exc import SkipTest

def test_latex():
    with runfilter('latex', LATEX) as doc:
        assert ".pdf" in doc.output_data().name
        assert doc.output_data().is_cached()

def test_latex_dvi():
    with runfilter('latexdvi', LATEX) as doc:
        assert ".dvi" in doc.output_data().name
        assert doc.output_data().is_cached()

def test_tikz():
    # TODO fix for tikz not installed
    raise SkipTest()
    with runfilter('tikz', TIKZ) as doc:
        assert ".pdf" in doc.output_data().name
        assert doc.output_data().is_cached()

def test_broken_latex():
    with wrap() as wrapper:
        wrapper.debug = False
        node = Doc("example.tex|latex",
                wrapper,
                [],
                contents = BROKEN_LATEX
                )
        wrapper.run_docs(node)
        assert wrapper.state == 'error'

TIKZ = """\
\\tikz \draw (0,0) -- (1,1)
{[rounded corners] -- (2,0) -- (3,1)}
-- (3,0) -- (2,1);
"""
LATEX = """\
\documentclass{article}
\\title{Hello, World!}
\\begin{document}
\maketitle
Hello!
\end{document}
"""

BROKEN_LATEX = """\
\documentclass{article}
"""



from dexy.utils import file_exists
import dexy.plugin
import os
import shutil
import sys

class Reporter(dexy.plugin.Plugin, metaclass=dexy.plugin.PluginMeta):
    """
    Base class for types of reporter.
    """

    _settings = {
            "default" : ("""Whether to run this report by default. Should be
            False for reports with side effects.""", True),
            "dir" : ("""Directory where report output will be written.""",
                None),
            'filename' : ("""Name of report file to generate (used when the
                report consists of just a single file).""", None),
            "in-cache-dir" : ("""Whether to create report dir in the cache
                directory (instead of project root).""", False),
            'no-delete' : ("""List of file names not to delete when resetting
                report dir (only effective if report dir is cleaned
                element-wise).""", ['.git', '.nojekyll']),
            'plugins' : ("""List of template plugin aliases which should be
                included in jinja environment.""",
                ['debug', 'inflection', 'builtins', 'operator', 'datetime',
                    'pprint', 'pygments', 'markdown']
                ),
            "run-for-wrapper-states" : ("""Wrapper states for which report is
                valid.""", ["ran"]),
            "readme-filename" : ("Name of README file, or None to omit." ,
                "README.md"),
            "readme-contents" : ("Contents to be written to README file." ,
                "This directory was generated by the %(alias)s dexy reporter and may be deleted without notice.\n"),
            "safety-filename" : ("""Name of a file which will be created in
                generated dir to indicate it was created by dexy and is safe to
                auto-remove.""", ".dexy-generated"), }

    def is_active(self):
        return True

    def help(self):
        """
        Allow implementing custom help since reporters can have such different behavior.
        """
        pass

    def copy_template_files(self):
        # Copy template files (e.g. .css files)
        # TODO shouldn't need to copy this each time - can just overwrite files which change
        class_file = sys.modules[self.__module__].__file__
        template_parent_dir = os.path.dirname(class_file)
        template_dir = os.path.join(template_parent_dir, 'files')
        shutil.copytree(template_dir, self.report_dir())
        self.write_safety_file()

    def run_plugins(self):
        env = {}
        for alias in self.setting('plugins'):
            from dexy.filters.templating_plugins import TemplatePlugin
            plugin = TemplatePlugin.create_instance(alias)
            self.log_debug("Running template plugin %s" % plugin.__class__.__name__)

            try:
                new_env_vars = plugin.run()
            except Exception:
                print(("error occurred processing template plugin '%s'" % alias))
                raise

            if any(v in list(env.keys()) for v in new_env_vars):
                new_keys = ", ".join(sorted(new_env_vars))
                existing_keys = ", ".join(sorted(env))
                msg = "plugin class '%s' is trying to add new keys '%s', already have '%s'"
                raise dexy.exceptions.InternalDexyProblem(msg % (plugin.__class__.__name__, new_keys, existing_keys))
            env.update(new_env_vars)
        return env

    def template_data(self):
        return dict((k, v[1]) for k, v in self.run_plugins().items())

    def cache_reports_dir(self):
        return os.path.join(self.wrapper.artifacts_dir, "reports")

    def report_dir(self):
        """
        Returns path of report directory relative to dexy project root.
        """
        if not self.setting('dir'):
            return None
        elif self.setting('in-cache-dir'):
            return os.path.join(self.cache_reports_dir(), self.setting('dir'))
        else:
            return self.setting('dir')

    def report_file(self):
        if self.setting('dir'):
            return os.path.join(self.report_dir(), self.setting('filename'))
        else:
            if self.setting('in-cache-dir'):
                return os.path.join(self.cache_reports_dir(), self.setting('filename'))
            else:
                return self.setting('dir')

    def key_for_log(self):
        return "reporter:%s" % self.aliases[0]

    def log_debug(self, message):
        self.wrapper.log.debug("%s: %s" % (self.key_for_log(), message))

    def log_info(self, message):
        self.wrapper.log.info("%s: %s" % (self.key_for_log(), message))

    def log_warn(self, message):
        self.wrapper.log.warn("%s: %s" % (self.key_for_log(), message))

    def safety_filepath(self):
        return os.path.join(self.report_dir(), self.setting('safety-filename'))

    def readme_filepath(self):
        readme_filename = self.setting('readme-filename')
        if readme_filename and readme_filename != 'None':
            return os.path.join(self.report_dir(), readme_filename)

    def write_safety_file(self):
        with open(self.safety_filepath(), "w") as f:
            f.write("""
            This directory was generated by the %s Dexy Reporter and
            may be deleted without notice.\n\n""" % self.__class__.__name__)

    def write_readme_file(self):
        with open(self.readme_filepath(), "w") as f:
            f.write(self.setting('readme-contents') % self.settings_and_attributes())

    def create_cache_reports_dir(self):
        if not file_exists(self.cache_reports_dir()):
            os.makedirs(self.cache_reports_dir())

    def create_reports_dir(self):
        if not self.report_dir():
            return False

        if not file_exists(self.report_dir()):
            os.makedirs(self.report_dir())

        self.write_safety_file()
        if self.readme_filepath():
            self.write_readme_file()

    def remove_reports_dir(self, wrapper, keep_empty_dir=False):
        self.wrapper = wrapper
        if not self.report_dir():
            return False

        if file_exists(self.report_dir()) and not file_exists(self.safety_filepath()):
            msg = """Please remove directory %s, Dexy wants to write a report
            here but there's already a file or directory in this location."""
            msgargs = (os.path.abspath(self.report_dir()),)
            raise dexy.exceptions.UserFeedback(msg % msgargs)
        elif file_exists(self.report_dir()):
            if keep_empty_dir:
                # Does not remove the base directory, useful if you are running
                # a process (like 'dexy serve') from inside that directory
                for f in os.listdir(self.report_dir()):
                    if not f in self.setting('no-delete'):
                        path = os.path.join(self.report_dir(), f)
                        if os.path.isdir(path):
                            shutil.rmtree(path)
                        else:
                            os.remove(path)
                self.write_safety_file()
            else:
                shutil.rmtree(self.report_dir())

    def run(self, wrapper):
        pass



from dexy.filter import DexyFilter
from dexy.utils import indent
from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.formatters import LatexFormatter
from pygments.formatters import get_formatter_for_filename
from pygments.lexers import LEXERS as PYGMENTS_LEXERS
from pygments.lexers import get_lexer_by_name
import dexy.commands
import dexy.exceptions
import posixpath
import pygments.lexers.web

pygments_lexer_cache = {}

file_ext_to_lexer_alias_cache = {
        '.pycon' : 'pycon',
        '.rbcon' : 'rbcon',
        '.Rd' : 'latex',
        '.svg' : 'xml',
        '.jinja' : 'jinja'
        }

# Add all pygments standard mappings.
for module_name, name, aliases, file_extensions, _ in list(PYGMENTS_LEXERS.values()):
    alias = aliases[0]
    for ext in file_extensions:
        ext = ext.lstrip("*")
        file_ext_to_lexer_alias_cache[ext] = alias

class SyntaxHighlightRstFilter(DexyFilter):
    """
    Surrounds code with highlighting instructions for ReST
    """
    aliases = ['pyg4rst']
    _settings = {
            'n' : ("Number of chars to indent.", 2),
            'data-type' : 'sectioned'
            }

    def process(self):
        n = self.setting('n')
        lexer_alias = file_ext_to_lexer_alias_cache[self.input_data.ext]

        for section_name, section_input in self.input_data.items():
            with_spaces = indent(section_input, n)
            section_output = ".. code:: %s\n\n%s" % (lexer_alias, with_spaces)
            self.output_data[section_name] = section_output

        self.output_data.save()

class SyntaxHighlightAsciidoctor(DexyFilter):
    """
    Surrounds code with highlighting instructions for Asciidoctor
    """
    aliases = ['asciisyn']
    _settings = {
            'lexer' : ("Specify lexer if can't be detected fro mfilename.", None),
            'data-type' : 'sectioned'
            }

    def process(self):
        if self.setting('lexer'):
            lexer_alias = self.setting('lexer')
        elif self.prev_filter and self.prev_filter.alias == 'idio':
            lexer_alias = file_ext_to_lexer_alias_cache[self.prev_filter.prev_ext]
        else:
            lexer_alias = file_ext_to_lexer_alias_cache[self.input_data.ext]

        for section_name, section_input in self.input_data.items():
            section_output = "[source,%s]\n----\n%s\n----\n" % (lexer_alias, section_input)
            self.output_data[section_name] = section_output

        self.output_data.save()

class PygmentsFilter(DexyFilter):
    """
    Apply Pygments <http://pygments.org/> syntax highlighting.
    
    Image output formats require PIL to be installed.
    """
    aliases = ['pyg', 'pygments']
    IMAGE_OUTPUT_EXTENSIONS = ['.png', '.bmp', '.gif', '.jpg']
    MARKUP_OUTPUT_EXTENSIONS = [".html", ".tex", ".svg", ".txt"] # make sure .html is first so it is default output format
    LEXER_ERR_MSG = """Pygments doesn't know how to syntax highlight files like '%s' (for '%s').\
    You might need to specify the lexer manually."""

    _settings = {
            'examples' : ['pygments', 'pygments-image', 'pygments-stylesheets'],
            'input-extensions' : [".*"],
            'output-extensions' : MARKUP_OUTPUT_EXTENSIONS + IMAGE_OUTPUT_EXTENSIONS + ['.css', '.sty'],

            'lexer' : ("""The name of the pygments lexer to use (will normally
            be determined automatically, only use this if you need to override
            the default setting or your filename isn't mapped to the lexer you
            want to use.""", None),
            'allow-unknown-ext' : ("""Whether to allow unknown file extensions
                to be parsed with the TextLexer by default instead of raising
                an exception.""", True),
            'allow-unprintable-input' : ("""Whether to allow unprintable input
                to be replaced with dummy text instead of raising an exception.""",
                True),
            'unprintable-input-text' : ("""Dummy text to use instead of
                unprintable binary input.""", 'not printable'),
            'lexer-args' : (
                "Dictionary of custom arguments to be passed directly to the lexer.",
                {}
                ),
            'lexer-settings' : (
                "List of all settings which will be passed to the lexer constructor.",
                []
            ),
            'formatter-settings' : (
                """List of all settings which will be passed to the formatter
                constructor.""", ['style', 'full', 'linenos', 'noclasses']
            ),

            'style' : ( "Formatter style to output.", 'default'),
            'noclasses' : ( "If set to true, token <span> tags will not use CSS classes, but inline styles.", None),
            'full' : ("""Pygments formatter option: output a 'full' document
                including header/footer tags.""", None),
            'linenos' : ("""Whether to include line numbers. May be set to
                'table' or 'inline'.""", None),
            'line-numbers' : ("""Alternative name for 'linenos'.""", None),
            }

    lexer_cache = {}

    def data_class_alias(klass, file_ext):
        if file_ext in klass.MARKUP_OUTPUT_EXTENSIONS:
            return 'sectioned'
        else:
            return 'generic'

    def docmd_css(klass, style='default'):
        """
        Prints out CSS for the specified style.
        """
        print(klass.generate_css(style))

    def docmd_sty(klass, style='default'):
        """
        Prints out .sty file (latex) for the specified style.
        """
        print(klass.generate_sty(style))

    def generate_css(self, style='default'):
        formatter = HtmlFormatter(style=style)
        return formatter.get_style_defs()

    def generate_sty(self, style='default'):
        formatter = LatexFormatter(style=style)
        return formatter.get_style_defs()

    def calculate_canonical_name(self):
        ext = self.prev_ext
        if ext in [".css", ".sty"] and self.ext == ext:
            return self.doc.name
        elif self.alias == 'htmlsections':
            name_without_ext = posixpath.splitext(self.doc.name)[0]
            return "%s%s" % (name_without_ext, self.ext)
        else:
            return "%s%s" % (self.doc.name, self.ext)

    def constructor_args(self, constructor_type, custom_args=None):
        if custom_args:
            args = custom_args
        else:
            args = {}

        for argname in self.setting("%s-settings" % constructor_type):
            if self.setting(argname):
                args[argname] = self.setting(argname)
        return args

    def lexer_alias(self, ext):
        if self.setting('lexer'):
            self.log_debug("custom lexer %s specified" % self.setting('lexer'))
            return self.setting('lexer')

        is_json_file = ext in ('.json', '.dexy') or self.output_data.name.endswith(".dexy")

        if is_json_file and (pygments.__version__ < '1.5'):
            return "javascript"
        elif is_json_file:
            return "json"

        if ext == '.Makefile' or (ext == '' and 'Makefile' in self.input_data.name):
            return 'makefile'

        try:
            return file_ext_to_lexer_alias_cache[ext]
        except KeyError:
            pass

    def create_lexer_instance(self):
        ext = self.prev_ext
        lexer_alias = self.lexer_alias(ext)
        lexer_args = self.constructor_args('lexer')
        lexer_args.update(self.setting('lexer-args'))

        if not lexer_alias:
            msg = self.LEXER_ERR_MSG
            msgargs = (self.input_data.name, self.key)

            if self.setting('allow-unknown-ext'):
                self.log_warn(msg % msgargs)
                lexer_alias = 'text'
            else:
                raise dexy.exceptions.UserFeedback(msg % msgargs)

        if lexer_alias in pygments_lexer_cache and not lexer_args:
            return pygments_lexer_cache[lexer_alias]
        else:
            lexer = get_lexer_by_name(lexer_alias, **lexer_args)
            if not lexer_args:
                pygments_lexer_cache[lexer_alias] = lexer
            return lexer

        return lexer

    def create_formatter_instance(self):
        if self.setting('line-numbers') and not self.setting('linenos'):
            self.update_settings({'linenos' : self.setting('line-numbers')})

        formatter_args = self.constructor_args('formatter', {
            'lineanchors' : self.output_data.web_safe_document_key() })
        self.log_debug("creating pygments formatter with args %s" % (formatter_args))

        return get_formatter_for_filename(self.output_data.name, **formatter_args)

    def process(self):
        if self.ext in self.IMAGE_OUTPUT_EXTENSIONS:
            try:
                import PIL
                PIL # because pyflakes
            except ImportError:
                print("python imaging library is required by pygments to create image output")
                raise dexy.exceptions.InactivePlugin('pyg')

        ext = self.prev_ext
        if ext in [".css", ".sty"] and self.ext == ext:
            # Special case if we get a virtual empty file, generate style file

            self.log_debug("creating a style file in %s" % self.key)
            if ext == '.css':
                output = self.generate_css(self.setting('style'))
            elif ext == '.sty':
                output = self.generate_sty(self.setting('style'))
            else:
                msg = "pyg filter doesn't know how to generate a stylesheet for %s extension"
                msgargs = (ext)
                raise dexy.commands.UserFeedback(msg % msgargs)

            self.output_data.set_data(output)
            self.update_all_args({'override-workspace-exclude-filters' : True })

        else:
            lexer = self.create_lexer_instance()

            if self.ext in self.IMAGE_OUTPUT_EXTENSIONS:
                # Place each section into an image.
                for k, v in self.input_data.items():
                    formatter = self.create_formatter_instance()
                    output_for_section = highlight(str(v).decode("utf-8"), lexer, formatter)
                    new_doc_name = "%s--%s%s" % (self.doc.key.replace("|", "--"), k, self.ext)
                    self.add_doc(new_doc_name, output_for_section)

                # Place entire contents into main file.
                formatter = self.create_formatter_instance()
                self.update_all_args({'override-workspace-exclude-filters' : True })
                with open(self.output_filepath(), 'wb') as f:
                    f.write(highlight(str(self.input_data), lexer, formatter))

            else:
                formatter = self.create_formatter_instance()
                for section_name, section_input in self.input_data.items():
                    try:
                        if isinstance(section_input, str):
                            section_input = section_input
                        else:
                            # If section_input is an instance of
                            # SectionValue then calling 'str' on this
                            # instance will call the __str__ method of
                            # SectionValue.
                            section_input = str(section_input)
                        section_output = highlight(section_input, lexer, formatter)
                    except UnicodeDecodeError:
                        if self.setting('allow-unprintable-input'):
                            section_input = self.setting('unprintable-input-text')
                            section_output = highlight(section_input, lexer, formatter)
                        else:
                            raise
                    self.output_data[section_name] = section_output
                self.output_data.save()




from dexy.filter import DexyFilter
from dexy.utils import parse_yaml
import re

class YamlargsFilter(DexyFilter):
    """
    Specify attributes in YAML at top of file.
    """
    aliases = ['yamlargs']

    def process_text(self, input_text):
        regex = "\r?\n---\r?\n"
        if re.search(regex, input_text):
            self.log_debug("Found yaml content.")
            raw_yamlargs, content = re.split(regex, input_text)
            yamlargs = parse_yaml(raw_yamlargs)
            self.log_debug("Adding yaml: %s" % yamlargs)
            self.add_runtime_args(yamlargs)
            return content
        else:
            self.log_debug("No yaml content found.")
            return input_text



from dexy.commands.utils import dummy_wrapper
import dexy.doc
import dexy.filter

def env_command():
    """
    Prints list of template plugins.
    """
    f = dexy.filter.Filter.create_instance("jinja")
    f.doc = dexy.doc.Doc('dummy', dummy_wrapper())
    jinja_template_filters = list(f.jinja_template_filters().keys())

    env = f.run_plugins()
    for k in sorted(env):
        try:
            helpstring, value = env[k]
        except Exception:
            print(k)
            print("Values should be a (docstring, value) tuple.")
            raise

        if k in jinja_template_filters:
            print("*%s: %s" % (k, helpstring,))
        else:
            print("%s: %s" % (k, helpstring,))

    print('')
    print("* indicates the method can be used as a jinja template filter")
    print('')

def plugins_command():
    """
    Prints list of plugin-able classes.
    """
    for plugin_class in sorted(dexy.plugin.Plugin.__subclasses__()):
        print(plugin_class.__name__)

from pygments import highlight
from pygments.lexers import PythonLexer
import dexy.data
import dexy.exceptions
import inspect
import pygments.formatters
from dexy.utils import indent
import textwrap

def datas_command(
        alias=False, # Alias of data type to print detaile dinfo for.
        source=False, # Whether to print source code for methods.
        nocolor=False, # If printing source, whether to colorize it.
    ):
    """
    Prints list of data types.
    """
    wrapper = dummy_wrapper()
    settings = {
            'canonical-name' : 'foo'
            }

    nodoc_methods = ('clear_cache', 'clear_data', 'copy_from_file', 'data', 'has_data',
            'initialize_settings', 'initialize_settings_from_other_classes',
            'initialize_settings_from_parents', 'initialize_settings_from_raw_kwargs',
            'is_active', 'is_cached', 'args_to_data_init', 'json_as_dict', 'as_text',
            'load_data', 'save', 'setup', 'setup_storage', 'storage_class_alias',
            'transition', 'add_to_lookup_sections' ,'add_to_lookup_nodes'
            )

    print("")

    if not alias:
        for d in dexy.data.Data.__iter__("foo", ".txt", "foo", settings, wrapper):
            print(d.alias)

        print("")
        print("For more information about a particular data type,")
        print("use the -alias flag and specify the data type alias.")
        print("")
    else:
        d = dexy.data.Data.create_instance(alias, "foo", ".txt", "foo", settings, wrapper)

        print(alias)
        print("")
        print(d.setting('help'))
        print("")

        print("Methods:")
        for k, v in inspect.getmembers(d):
            if k.startswith('_'):
                continue

            if inspect.ismethod(v) and not k in nodoc_methods:
                print("    %s" % k)

                docs = inspect.getdoc(v)
                if not docs:
                    raise dexy.exceptions.InternalDexyProblem("Must provide docstring for %s" % k)

                print("")
                print(indent(docs, 8))
                print("")

                args, varargs, keywords, defaults = inspect.getargspec(v)

                if not source and len(args) > 1:
                    print("        Takes arguments. Run with -source option to see source code.")


                if source:
                    source_code = textwrap.dedent(inspect.getsource(v))
                    if nocolor:
                        print(indent(source_code, 8))
                    else:
                        formatter = pygments.formatters.TerminalFormatter()
                        lexer = PythonLexer()
                        print(indent(highlight(source_code, lexer, formatter), 8))

                print("")



from dexy.doc import Doc
from tests.utils import wrap
import os

def test_output_name_with_filters():
    with wrap() as wrapper:
        doc = Doc(
                "data.md|markdown",
                wrapper,
                [],
                output_name="finished.bar",
                contents="foo"
                )
        wrapper.run_docs(doc)

        assert doc.initial_data.name == "data.md"
        assert doc.output_data().output_name() == "finished.bar"

def test_custom_name():
    with wrap() as wrapper:
        doc = Doc(
                "foo/data.txt",
                wrapper,
                [],
                output_name="data.abc",
                contents="12345.67"
                )
        wrapper.run_docs(doc)

        assert doc.output_data().output_name() == "foo/data.abc"

def test_custom_name_in_subdir():
    with wrap() as wrapper:
        doc = Doc(
                "data.txt",
                wrapper,
                [],
                output_name="subdir/data.abc",
                contents="12345.67"
                )
        wrapper.run_docs(doc)
        wrapper.report()

        assert doc.output_data().output_name() == "subdir/data.abc"
        assert doc.output_data().parent_output_dir() == "subdir"

def test_custom_name_with_args():
    with wrap() as wrapper:
        doc = Doc(
                "data.txt",
                wrapper,
                [],
                output_name="%(bar)s/data-%(foo)s.abc",
                foo='bar',
                bar='baz',
                contents="12345.67"
                )
        wrapper.run_docs(doc)
        wrapper.report()

        assert doc.output_data().output_name() == "baz/data-bar.abc"
        assert doc.output_data().parent_output_dir() == "baz"
        assert os.path.exists("output/baz/data-bar.abc")

def test_custom_name_with_leading_slash():
    with wrap() as wrapper:
        doc = Doc(
                "foobarbaz/data.txt",
                wrapper,
                [],
                output_name="/%(bar)s/data-%(foo)s.abc",
                foo='bar',
                bar='baz',
                contents="12345.67"
                )
        wrapper.run_docs(doc)
        assert doc.output_data().output_name() == "baz/data-bar.abc"



from datetime import datetime
from dexy.filters.api import ApiFilter
import dexy.exceptions
import getpass
import os
import urllib

try:
    import boto
    from boto.s3.key import Key
    BOTO_AVAILABLE = True
except ImportError:
    BOTO_AVAILABLE = False

class BotoUploadFilter(ApiFilter):
    """
    Uses boto library to upload content to S3, returns the URL.

    You can set AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY variables in your
    system environment (the environment that runs the dexy command) or you can
    set defaults in your ~/.dexyapis file (these will override the
    environment):

    "AWS" : {
        "AWS_ACCESS_KEY_ID" : "AKIA...",
        "AWS_SECRET_ACCESS_KEY" : "hY6cw...",
        "AWS_BUCKET_NAME" : "my-unique-bucket-name"
    }

    You can also have a .dexyapis file in the directory in which you run Dexy,
    and this will override the user-wide .dexyapis file. You can use this to
    specify a per-project bucket.

    You can add a date to your bucket by specifying strftime codes in your
    bucket name, this is useful so you don't have to worry about all your
    filenames being unique.

    If you do not set bucket-name, it will default to a name based on your
    username. This may not be unique across all S3 buckets so it may be
    necessary for you to specify a name. You can use an existing S3 bucket,
    a new bucket will be created if your bucket does not already exist.
    """
    aliases = ['s3', 'botoup']
    _settings = {
            'api-key-name' : 'AWS',
            'output-extensions' : ['.txt'],
            }
    API_KEY_KEYS = ['AWS_ACCESS_KEY_ID', 'AWS_SECRET_ACCESS_KEY', 'AWS_BUCKET_NAME']

    def is_active(self):
        return BOTO_AVAILABLE

    def bucket_name(self):
        """
        Figure out which S3 bucket name to use and create the bucket if it doesn't exist.
        """
        bucket_name = self.read_param('AWS_BUCKET_NAME')
        if not bucket_name:
            try:
                username = getpass.getuser()
                bucket_name = "dexy-%s" % username
                return bucket_name
            except dexy.exceptions.UserFeedback:
                print("Can't automatically determine username. Please specify AWS_BUCKET_NAME for upload to S3.")
                raise
        bucket_name = datetime.now().strftime(bucket_name)
        self.log_debug("S3 bucket name is %s" % bucket_name)
        return bucket_name

    def boto_connection(self):
        if os.getenv('AWS_ACCESS_KEY_ID') and os.getenv('AWS_SECRET_ACCESS_KEY'):
            # use values defined in env
            return boto.connect_s3()
        else:
            # use values specified in .dexyapis
            aws_access_key_id = self.read_param('AWS_ACCESS_KEY_ID')
            aws_secret_access_key = self.read_param('AWS_SECRET_ACCESS_KEY')
            return boto.connect_s3(aws_access_key_id, aws_secret_access_key)

    def get_bucket(self):
        conn = self.boto_connection()
        return conn.create_bucket(self.bucket_name())

    def process(self):
        b = self.get_bucket()
        k = Key(b)
        k.key = self.input_data.web_safe_document_key()
        self.log_debug("Uploading contents of %s" % self.input_data.storage.data_file())
        k.set_contents_from_filename(self.input_data.storage.data_file())
        k.set_acl('public-read')
        url = "https://s3.amazonaws.com/%s/%s" % (self.bucket_name(), urllib.quote(k.key,))
        self.output_data.set_data(url)



import dexy.filters
import dexy.reporters
import dexy.parsers
import dexy.datas

# Automatically register plugins in any python package named like dexy_*
import pkg_resources
for dist in pkg_resources.working_set:
    if dist.key.startswith("dexy-"):
        import_pkg = dist.egg_name().split("-")[0]
        try:
            __import__(import_pkg)
        except ImportError as e:
            print(("plugin", import_pkg, "not registered because", e))



from dexy.data import Data
from dexy.doc import Doc
from dexy.exceptions import UserFeedback
from nose.tools import raises
from tests.utils import wrap

def test_create_doc_with_one_filter():
    with wrap() as wrapper:
        doc = Doc("foo.txt|dexy", wrapper, [], contents="foo")

        assert len(doc.filters) == 1
        f = doc.filters[0]

        assert f.doc == doc
        assert not f.prev_filter
        assert not f.next_filter

        wrapper.run_docs(doc)

def test_create_doc_with_two_filters():
    with wrap() as wrapper:
        doc = Doc("foo.txt|dexy|dexy", wrapper, [], contents="foo")
        assert len(doc.filters) == 2

        f1, f2 = doc.filters

        assert f1.doc == doc
        assert f2.doc == doc

        assert f1.next_filter == f2
        assert f2.prev_filter == f1

        assert not f1.prev_filter
        assert not f2.next_filter

@raises(UserFeedback)
def test_blank_alias():
    with wrap() as wrapper:
        Doc("abc.txt|", wrapper, [], contents='foo')

def test_output_is_data():
    with wrap() as wrapper:
        doc = Doc("abc.txt", wrapper, [], contents="these are the contents")
        wrapper.run_docs(doc)
        assert isinstance(doc.output_data(), Data)

def test_create_virtual_initial_artifact():
    with wrap() as wrapper:
        doc = Doc("abc.txt", wrapper, [], contents="these are the contents")
        wrapper.run_docs(doc)
        assert doc.output_data().__class__.__name__ == "Generic"



from dexy.filter import DexyFilter
import json
import re

class MarkdownSections(DexyFilter):
    """
    Base class for filters which divide markdown scripts into code/prose
    sections.
    """

    aliases = ['mdsections']
    _settings = {
            "input-extensions" : [".md"],
            "output-extensions" : [".json"],
            "output" : True,
            "language" : ("Default programming language for code blocks.", "python"),
            "pprint"  : ("Whether to pretty print JSON.", True),
            }


    def process_code(self, source, language, metadata=None):
        return {
                "type" : "source",
                "language" : language,
                "source" : source,
                "metadata" : metadata
                }

    def process_heading(self, level, text, metadata=None):
        return {
                "type" : "heading",
                "level" : level,
                "text" : text,
                "metadata" : metadata
                }

    def process_prose(self, source, metadata=None):
        return {
                "type" : "markdown",
                "text" : source,
                "metadata" : metadata
                }

    def process_sections(self, input_text):
        blocks = []

        proseblock = []
        codeblock = None

        language = None
        state = "md"

        for line in input_text.splitlines():
            if state == "md" and line.lstrip().startswith("```"):
                # save exististing prose block, if any
                if proseblock:
                    block = self.process_prose(proseblock)
                    blocks.append(block)
                    proseblock = []

                # start new code block, skipping current line
                state = "code"

                # Detect lexer, if specified
                match_lexer = re.match("```([A-Za-z-]+)", line)
                if match_lexer:
                    language = match_lexer.groups()[0]
                else:
                    language = self.setting('language')

                codeblock = []
            elif state == "code" and line.lstrip().startswith("```"):
                block = self.process_code(codeblock, language)
                blocks.append(block)

                state = "md"
            elif state == "code":
                codeblock.append(line)
            elif state == "md":
                m = re.match("^(#+)(\s*)(.*)$", line)
                if m:
                    if proseblock:
                        block = self.process_prose(proseblock)
                        blocks.append(block)
                        proseblock = []

                    level = len(m.groups()[0])
                    block = self.process_heading(level, m.groups()[2])
                    blocks.append(block)
                else:
                    proseblock.append(line)

        return blocks
        
    def process_text(self, input_text):
        blocks = self.process_sections(input_text)

        if self.setting('pprint'):
            return json.dumps(blocks, indent=4, sort_keys=True)
        else:
            return json.dumps(blocks)


class MarkdownJupyter(MarkdownSections):
    """
    Generate a Jupyter (IPython) notebook from markdown with embedded code.
    """
    aliases = ['mdipynb', 'mdjup']
    _settings = {
            "output-extensions" : [".ipynb"],
            "nbformat" : ("Setting to use for IPython nbformat setting", 3),
            "nbformat_minor" : ("Setting to use for IPython nbformat_minor setting.", 0),
            "name" : ("Name of notebook.", None),
            "collapsed" : ("Whether to collapse code blocks by default.", False),
            }
    
    def process_code(self, source, language, metadata=None):
        if language is None:
            raise Exception("no language specified")
        return {
                "cell_type" : "code",
                "collapsed" : self.setting('collapsed'),
                "metadata" : metadata or {},
                "language" : language,
                "input" : source,
                "outputs" : [],
                "prompt_number" : None
                }

    def process_heading(self, level, text, metadata = None):
        return {
                "cell_type" : "heading",
                "level" : level,
                "metadata" : metadata or {},
                "source" : [text]
                }

    def process_prose(self, source):
        return {
                "cell_type" : "markdown",
                "metadata" : {},
                "source" : [
                    "\n".join(source)
                    ]
                }

    def process_text(self, input_text):
        workbook_name = self.setting("name")
        cells = self.process_sections(input_text)

        notebook = {
            "nbformat" : self.setting('nbformat'),
            "nbformat_minor" : self.setting('nbformat-minor'),
            "metadata" : {
                "name" : workbook_name
                },
            "worksheets" : [{
                "cells" : cells
                }]
            }

        if self.setting('pprint'):
            return json.dumps(notebook, indent=4, sort_keys=True)
        else:
            return json.dumps(notebook)



import dexy.filter
import inspect

def fcmds_command(
        alias=False # Only print commands defined by this alias.
        ):
    """
    Prints a list of available filter commands.
    """
    if alias:
        filter_instances = [dexy.filter.Filter.create_instance(alias)]
    else:
        filter_instances = dexy.filter.Filter

    for filter_instance in filter_instances:
        cmds = filter_instance.filter_commands()
        if cmds:
            print("filter alias:", filter_instance.alias)
            for command_name in sorted(cmds):
                docs = inspect.getdoc(cmds[command_name])
                if docs:
                    doc = docs.splitlines()[0]
                    print("    %s   # %s" % (command_name, doc))
                else:
                    print("    %s" % command_name)
            print('')

def fcmd_command(
        alias=None, # The alias of the filter which defines the custom command
        cmd=None, # The name of the command to run
        **kwargs # Additional arguments to be passed to the command
        ):
    """
    Run a filter command.
    """
    filter_instance = dexy.filter.Filter.create_instance(alias)
    cmd_name = "docmd_%s" % cmd

    if not cmd_name in dir(filter_instance):
        msg = "%s is not a valid command. There is no method %s defined in %s"
        msgargs = (cmd, cmd_name, filter_instance.__class__.__name__)
        raise dexy.exceptions.UserFeedback(msg % msgargs)

    else:
        instance_method = getattr(filter_instance, cmd_name)
        # TODO use try/catch instead of inspect.ismethod
        if inspect.ismethod(instance_method):
            try:
                instance_method.__func__(filter_instance, **kwargs)
            except TypeError as e:
                print(e.message)
                print(inspect.getargspec(instance_method.__func__))
                print(inspect.getdoc(instance_method.__func__))
                raise

        else:
            msg = "expected %s to be an instance method of %s"
            msgargs = (cmd_name, filter_instance.__class__.__name__)
            raise dexy.exceptions.InternalDexyProblem(msg % msgargs)



from dexy.exceptions import UserFeedback
from dexy.filters.process import SubprocessFilter
import os

class CasperJsSvg2PdfFilter(SubprocessFilter):
    """
    Converts an SVG file to PDF by running it through casper js.

    # TODO convert this to phantomjs, no benefit to using casper here (js is
    # not user facing) and more restrictive
    """
    aliases = ['svg2pdf']
    _settings = {
            'add-new-files' : True,
            'executable' : 'casperjs',
            'version-command' : 'casperjs --version',
            "input-extensions" : ['.svg'],
            "output-extensions" : ['.pdf'],
            "width" : ("Width of page to capture.", 200),
            "height" : ("Height of page to capture.", 200),
            "command-string" : "%(prog)s %(args)s script.js"
            }

    def script_js(self, width, height):
        args = {
                'width' : width,
                'height' : height,
                'svgfile' : self.work_input_filename(),
                'pdffile' : self.work_output_filename()
                }
        return """
        var casper = require('casper').create({
             viewportSize : {width : %(width)s, height : %(height)s}
        });
        casper.start('%(svgfile)s', function() {
            this.capture('%(pdffile)s');
        });

        casper.run();
        """ % args

    def custom_populate_workspace(self):
        width = self.setting('width')
        height = self.setting('height')
        js = self.script_js(width, height)

        wd = self.parent_work_dir()
        scriptfile = os.path.join(wd, "script.js")

        self.log_debug("scriptfile: %s" % scriptfile)
        self.log_debug("js for scriptfile: %s" % js)

        with open(scriptfile, "w") as f:
            f.write(js)

class PhantomJsRenderSubprocessFilter(SubprocessFilter):
    """
    Renders HTML to PNG/PDF using phantom.js.
    
    If the HTML relies on local assets such as CSS or image files, these should
    be specified as inputs.

    See phantomjs documentation for paper sizing options.
    http://phantomjs.org/api/webpage/property/paper-size.html
    """
    aliases = ['phrender']
    _settings = {
            'add-new-files' : True,
            'examples' : ['phrender'],
            'executable' :  'phantomjs',
            "paper-size" : ("Paper size (e.g. phantomjs 'pageSize.format' setting).", "A4"),
            "orientation" : ("e.g. landscape or portrait orientation.", None),
            "page-width" : ("Paper width for page.paperSize phantom setting.", None),
            "page-height" : ("Paper height for page.paperSize phantom setting.", None),
            "border" : ("Border around page.", None),
            "page-header-height" : ("Height of header to print on each page.", None),
            "page-header-contents" : ("Custom HTML header to print on each page.", None),
            "page-footer-height" : ("Height of footer to print on each page.", None),
            "page-footer-contents" : ("Custom HTML footer to print on each page.", None),
            'version-command' : 'phantomjs --version',
            'command-string' : "%(prog)s %(args)s script.js",
            'input-extensions' : [".html", ".htm", ".txt"],
            'output-extensions' : [".png", ".pdf"]
            }

    def custom_populate_workspace(self):
        timeout = self.setup_timeout()
        if not timeout:
            raise Exception("must have timeout")

        args = {
                'address' : self.work_input_filename(),
                'output' : self.work_output_filename(),
                'timeout' : timeout,
                }

        page_width = self.setting('page-width')
        page_height = self.setting('page-height')
        border = self.setting('border')
        orientation = self.setting('orientation')

        if page_width is not None and page_height is not None:
            args['paper_size'] = "width: '%s',\nheight: '%s',\n" % (page_width, page_height)
        else:
            args['paper_size'] = "format: '%s',\n" % format
            if orientation is not None:
                args['paper_size'] += "orientation: '%s',\n" % orientation

        if border is not None:
            args['paper_size'] += "border: '%s',\n" % border

        if self.setting('page-header-height') is not None:
            args['paper_size'] += """header: {
                height: "%s",
                contents: phantom.callback(function(pageNum, numPages) {
                    return %s
                })
                },
            """ % (self.setting('page-header-height'), self.setting('page-header-contents'),)

        if self.setting('page-footer-height') is not None:
            args['paper_size'] += """footer: {
                height: "%s",
                contents: phantom.callback(function(pageNum, numPages) {
                    return %s
                })
                },
            """ % (self.setting('page-footer-height'), self.setting('page-footer-contents'),)

        self.log_debug("args are: %s" % args)


        js = """
        address = '%(address)s'
        output = '%(output)s'
        var page = new WebPage(),
            address, output, size;

        page.paperSize = {
            %(paper_size)s
        };

        page.open(address, function (status) {
            if (status !== 'success') {
                console.log('Unable to load the address!');
            } else {
                window.setTimeout(function () {
                page.render(output);
                phantom.exit();
                }, %(timeout)s);
            }
        });
        """ % args

        wd = self.parent_work_dir()
        scriptfile = os.path.join(wd, "script.js")
        self.log_debug("scriptfile: %s" % scriptfile)
        self.log_debug("js for scriptfile: %s" % js)
        with open(scriptfile, "w") as f:
            f.write(js)



from dexy.filter import DexyFilter
from dexy.plugin import TemplatePlugin

# https://pypi.python.org/pypi/ansi2html
# https://github.com/ralphbean/ansi2html

try:
    from ansi2html import Ansi2HTMLConverter
    AVAILABLE = True
except ImportError:
    AVAILABLE = False

class Ansi2HTMLTemplatePlugin(TemplatePlugin):
    """
    Expose ansi2html within templates.
    """
    aliases = ['ansi2html']

    def is_active(self):
        return AVAILABLE

    def convert(self, doc, font_size='normal'):
        conv = Ansi2HTMLConverter(inline=True, font_size=font_size)
        return conv.convert(str(doc), full=False)

    def run(self):
        return { 'ansi2html' : ("The convert method from ansi2html module.", self.convert) }

class Ansi2HTMLFilter(DexyFilter):
    """
    Generates HTML from ANSI color codes using ansi2html.
    """
    aliases = ['ansi2html']
    _settings = {
            'output-extensions' : ['.html'],
            'input-extensions' : ['.txt', '.sh-session'],
            'data-type' : 'sectioned',
            'pre' : ("Whether to wrap in <pre> tags.", True),
            'font-size' : ("CSS font size to be used.", "normal")
            }

    def is_active(self):
        return AVAILABLE

    def process(self):
        conv = Ansi2HTMLConverter(inline=True, font_size=self.setting('font-size'))
        if self.setting('pre'):
            s = "<pre>\n%s</pre>\n"
        else:
            s = "%s\n"

        for k, v in self.input_data.items():
            self.output_data[k] = s % conv.convert(str(v), full=False)
        self.output_data.save()




from dexy.utils import copy_or_link
from dexy.utils import os_to_posix
from operator import attrgetter
import dexy.doc
import dexy.exceptions
import dexy.plugin
import dexy.utils
import os
import posixpath

class FilterException(Exception):
    pass

class Filter(dexy.plugin.Plugin, metaclass=dexy.plugin.PluginMeta):
    """
    Base class for types of filter.
    """

    TAGS = []
    _class_settings = {'max-docstring-length' : 75}
    nodoc_settings = [
            'help', 'nodoc'
            ]
    _settings = {
            'added-in-version' : (
                "Dexy version when this filter was first available.",
                ''),
            'add-new-files' : (
                "Boolean or list of extensions/patterns to match.",
                False),
            'exclude-add-new-files' : (
                "List of patterns to skip even if they match add-new-files.",
                []),
            'exclude-new-files-from-dir' : (
                "List of directories to skip when adding new files.",
                []),
            'additional-doc-filters' : (
                "Filters to apply to additional documents created as side effects.",
                {}),
            'additional-doc-settings' : (
                "Settings to apply to additional documents created as side effects.",
                {}),
            'examples' : (
                "Templates which should be used as examples for this filter.",
                []),
            'ext' : (
                'File extension to output.',
                None),
            'extension-map' : (
                "Dictionary mapping input extensions to default output extensions.",
                None),
            'help' : (
                'Help string for filter, if not already specified as a class docstring.',
                None),
            'input-extensions' : (
                "List of extensions which this filter can accept as input.",
                [".*"]),
            'keep-originals' : (
                """Whether, if additional-doc-filters are specified, the
                original unmodified docs should also be added.""",
                False),
            'mkdir' : (
                "A directory which should be created in working dir.",
                None),
            'mkdirs' : (
                "A list of directories which should be created in working dir.",
                []),
            'nodoc' : (
                "Whether filter should be excluded from documentation.",
                False),
            'output' : (
                """Whether to output results of this filter by default by
                reporters such as 'output' or 'website'.""",
                False),
            'data-type' : (
                "Alias of custom data class to use to store filter output.",
                "generic"),
            'output-extensions' : (
                "List of extensions which this filter can produce as output.",
                [".*"]),
            'preserve-prior-data-class' : (
                "Whether output data class should be set to match the input data class.",
                False),
            'require-output' : (
                "Should dexy raise an exception if no output is produced by this filter?",
                True),
            'tags' : (
                "Tags which describe the filter.",
                []),
            'variables' : (
                'A dictionary of variable names and values to make available to this filter.',
                {}),
            'vars' : (
                'A dictionary of variable names and values to make available to this filter.',
                {}),
            'workspace-exclude-filters' : (
                "Filters whose output should be excluded from workspace.",
                ['pyg']),
            'override-workspace-exclude-filters' : (
                """If True, document will be populated to other workspaces
                ignoring workspace-exclude-filters.""", False),
            'workspace-includes' : (
                """If set to a list of filenames or extensions, only these will
                be populated to working dir.""",
                None)
            }

    def __init__(self, doc=None):
        self.doc = doc

    def filter_commands(self):
        """
        Return dictionary of filter command canonical names and method objects.
        """
        fcmds = {}
        for m in dir(self):
            if m.startswith("docmd_"):
                key = m.replace("docmd_", "")
                fcmds[key] = getattr(self, m)
        return fcmds

    def final_ext(self):
        return self.doc.output_data().ext

    def data_class_alias(self, ext):
        if self.setting('preserve-prior-data-class'):
            return self.input_data.alias
        else:
            return self.setting('data-type')

    def add_runtime_args(self, new_args):
        self.doc.add_runtime_args(new_args)

    def update_all_args(self, new_args):
        self.doc.add_runtime_args(new_args)

    def setup(self, key, storage_key, prev_filter, next_filter, custom_settings):
        self.key = key
        self.storage_key = storage_key
        self.prev_filter = prev_filter
        self.next_filter = next_filter

        self.update_settings(custom_settings)

        if self.prev_filter:
            self.input_data = self.prev_filter.output_data
            self.prev_ext = self.prev_filter.ext
        else:
            self.input_data = self.doc.initial_data
            self.prev_ext = self.doc.initial_data.ext

        self.set_extension()

        settings = self.input_data.setting_values()

        del settings['storage-type']
        del settings['aliases']
        
        settings.update({
                'canonical-name' : self.calculate_canonical_name(),
                'canonical-output' : self.is_canonical_output()
                })

        self.output_data = dexy.data.Data.create_instance(
                self.data_class_alias(self.ext),
                self.key,
                self.ext,
                self.storage_key,
                settings,
                self.doc.wrapper
                )

    def is_canonical_output(self):
        if self.input_data.setting('canonical-output') == True:
            return True
        elif (self.input_data.setting('canonical-output') == False):
            return False
        elif self.setting('output'):
            return True
        else:
            return None

    def set_extension(self):
        i_accept = self.setting('input-extensions')
        i_output = self.setting('output-extensions')

        if self.prev_filter:
            prev_ext = self.prev_filter.ext
        else:
            prev_ext = self.doc.ext

        # Check that we can handle input extension
        if set([prev_ext, ".*"]).isdisjoint(set(i_accept)):
            msg = "Filter '%s' in '%s' can't handle file extension %s, supported extensions are %s"
            params = (self.alias, self.key, prev_ext, ", ".join(i_accept))
            raise dexy.exceptions.UserFeedback(msg % params)

        # Figure out output extension
        ext = self.setting('ext')
        if ext:
            # User has specified desired extension
            if not ext.startswith('.'):
                ext = '.%s' % ext

            # Make sure it's a valid one
            if (not ext in i_output) and (not ".*" in i_output):
                msg = "You have requested file extension %s in %s but filter %s can't generate that."
                raise dexy.exceptions.UserFeedback(msg % (ext, self.key, self.alias))

            self.ext = ext

        elif ".*" in i_output:
            self.ext = prev_ext

        else:
            # User has not specified desired extension, and we don't output wildcards,
            # figure out extension based on next filter in sequence, if any.
            ext_from_map = None
            if self.setting('extension-map'):
                ext_from_map = self.setting('extension-map')[prev_ext]

            next_filter_accepts = [".*"]
            if self.next_filter:
                next_filter_accepts = self.next_filter.setting('input-extensions')

            if ".*" in next_filter_accepts:
                if ext_from_map:
                    self.ext = ext_from_map
                else:
                    self.ext = i_output[0]

            elif ext_from_map:
                if not ext_from_map in next_filter_accepts:
                    msg = "Filter %s wants to output %s but %s doesn't accept this format."
                    msgargs = (self.alias, ext_from_map, self.next_filter.alias)
                    raise dexy.exceptions.UserFeedback(msg % msgargs)

                self.ext = ext_from_map

            else:
                if set(i_output).isdisjoint(set(next_filter_accepts)):
                    msg = "Filter %s can't go after filter %s, no file extensions in common."
                    raise dexy.exceptions.UserFeedback(msg % (self.next_filter.alias, self.alias))

                for e in i_output:
                    if e in next_filter_accepts:
                        self.ext = e

                if not self.ext:
                    msg = "no file extension found but checked already for disjointed, should not be here"
                    raise dexy.exceptions.InternalDexyProblem(msg)

    def templates(self):
        """
        List of dexy templates which refer to this filter.
        """
        import dexy.template
        templates = [dexy.template.Template.create_instance(a) for a in self.setting('examples')]
        return templates

    def key_with_class(self):
        return "%s:%s" % (self.__class__.__name__, self.key)

    def log_debug(self, message):
        self.doc.log_debug(message)

    def log_info(self, message):
        self.doc.log_info(message)

    def log_warn(self, message):
        self.doc.log_warn(message)

    def process(self):
        """
        Run the filter, converting input to output.
        """
        pass

    def calculate_canonical_name(self):
        name_without_ext = posixpath.splitext(self.doc.name)[0]
        return "%s%s" % (name_without_ext, self.ext)

    def output_filepath(self):
        return self.output_data.storage.data_file()

    def doc_arg(self, arg_name_hyphen, default=None):
        return self.doc.arg_value(arg_name_hyphen, default)

    def add_doc(self, doc_name, doc_contents=None, doc_args = None):
        """
        Creates a new Doc object for an on-the-fly document.
        """
        self.log_debug("adding doc with contents type %s" % doc_contents.__class__.__name__)
        doc_name = os_to_posix(doc_name)
        if not posixpath.sep in doc_name:
            doc_name = posixpath.join(self.input_data.parent_dir(), doc_name)

        doc_ext = os.path.splitext(doc_name)[1]

        additional_doc_filters = self.setting('additional-doc-filters')
        self.log_debug("additional-doc-filters are %s" % additional_doc_filters)

        
        additional_doc_settings = self.setting('additional-doc-settings')

        settings = None
        if isinstance(additional_doc_settings, list):
            # figure out which settings to apply based on file extension
            for pattern, settings in additional_doc_settings:
                if doc_ext == pattern or pattern == ".*":
                    break
        elif isinstance(additional_doc_settings, dict):
            settings = additional_doc_settings

        else:
            raise Exception("Unexpected type %s" % type(settings))

        if doc_args:
            settings.update(doc_args)

        self.log_debug("additional-doc-settings are %s" % settings)


        def create_doc(name, filters, contents, args=None):
            if filters:
                doc_key = "%s|%s" % (name, filters)
            else:
                doc_key = name

            if not args:
                args = {}

            doc = dexy.doc.Doc(
                    doc_key,
                    self.doc.wrapper,
                    [],
                    contents=contents,
                    **args
                    )

            doc.output_data().setup()
            doc.output_data().storage.connect()

            self.doc.add_additional_doc(doc)
            return doc

        doc = None

        if isinstance(additional_doc_filters, str):
            doc = create_doc(doc_name,
                    additional_doc_filters, doc_contents, settings)

        elif isinstance(additional_doc_filters, list):
            for f in reversed(additional_doc_filters):
                doc = create_doc(doc_name, f, doc_contents, settings)

        elif isinstance(additional_doc_filters, dict):
            filters = additional_doc_filters.get(doc_ext)
            if isinstance(filters, list):
                for f in filters:
                    doc = create_doc(doc_name, f, doc_contents, settings)
            elif isinstance(filters, str):
                doc = create_doc(doc_name, filters, doc_contents, settings)
            elif filters is None:
                pass
            else:
                msg = "additional_doc_filters values should be list of string. Received %s"
                msgargs = filters.__class__.__name__
                raise Exception(msg % msgargs)

        else:
            msg = "additional-doc-filters should be string, list or dict. Received %s"
            msgargs = additional_doc_filters.__class__.__name__
            raise dexy.exceptions.InternalDexyProblem(msg % msgargs)

        if self.setting('keep-originals') or doc is None:
            doc = create_doc(doc_name, '', doc_contents, settings)

        return doc

    def workspace(self):
        """
        Directory in which all working files for this filter are stored.

        The `populate_workspace` method will populate this directory with
        inputs to this filter.
        """
        ws = self.doc.wrapper.work_cache_dir()
        return os.path.join(ws, self.storage_key[0:2], self.storage_key)

    def parent_work_dir(self):
        """
        Within the 'workspace', this is the parent directory of the file to be
        processed. This is the directory which subprocess/pexpect will 'cwd' to
        and execute processes.
        """
        return os.path.join(self.workspace(), self.output_data.parent_dir())

    def work_input_filename(self):
        """
        Name of work file to use input from. Processes will take this file name
        as input. Does not contain full path to file, just the file name. File
        will be in parent_work_dir() and processes should set their working
        directory to parent_work_dir() before running.
        """
        if self.ext and (self.ext == self.prev_ext):
            return "%s-work%s" % (self.input_data.baserootname(), self.prev_ext)
        else:
            return self.input_data.basename()

    def work_input_filepath(self):
        return os.path.join(self.parent_work_dir(), self.work_input_filename())

    def work_output_filename(self):
        """
        Name of work file to save output to. Processes will take this file name
        as output. Does not contain full path to file, just the file name. File
        will be in parent_work_dir() and processes should set their working
        directory to parent_work_dir() before running.
        """
        return self.output_data.basename()

    def work_output_filepath(self):
        return os.path.join(self.parent_work_dir(), self.work_output_filename())

    def include_input_in_workspace(self, inpt):
        """
        Whether to include the contents of the input file inpt in the workspace
        for this filter.
        """
        workspace_includes = self.setting('workspace-includes')

        if workspace_includes is not None:
            if inpt.ext in workspace_includes:
                self.log_debug("Including %s because file extension matches." % inpt)
                return True
            elif inpt.output_data().basename() in workspace_includes:
                self.log_debug("Including %s because base name matches." % inpt)
                return True
            else:
                self.log_debug("Excluding %s because does not match workspace-includes" % inpt)
                return False

        elif not inpt.filters:
            self.log_debug("Including because %s has no filters." % inpt)
            return True

        elif inpt.filters[-1].setting('override-workspace-exclude-filters'):
            self.log_debug("Including %s because override-workspace-exclude-filters is set." % inpt)
            return True

        else:
            workspace_exclude_filters = self.setting('workspace-exclude-filters')

            if workspace_exclude_filters is None:
                self.log_debug("Including because exclude_filters is None.")
                return True
            elif any(a in workspace_exclude_filters for a in inpt.filter_aliases):
                self.log_debug("Excluding %s because of workspace-exclude-filters" % inpt)
                return False
            else:
                self.log_debug("Including %s because not excluded" % inpt)
                return True

    def makedirs(self):
        mkdirs = self.setting('mkdirs')

        # mkdir should be a string, but handle either string or list
        mkdir = self.setting('mkdir')
        if mkdir:
            if isinstance(mkdir, str):
                mkdirs.append(mkdir)
            else:
                mkdirs.extend(mkdir)

        for d in mkdirs:
            dirpath = os.path.join(self.workspace(), d)
            self.log_debug("Creating directory %s" % dirpath)
            os.makedirs(dirpath)

    def populate_workspace(self):
        """
        Populates the workspace directory with inputs to the filter, under
        their canonical names.
        """
        self.log_debug("in populate_workspace method")
        already_created_dirs = set()
        wd = self.parent_work_dir()

        self._files_workspace_populated_with = set()

        self.doc.wrapper.trash(wd)

        try:
            os.makedirs(wd)
            already_created_dirs.add(wd)
        except OSError:
            msg = "workspace '%s' for filter '%s' already exists"
            msgargs = (os.path.abspath(wd), self.key,)
            raise dexy.exceptions.InternalDexyProblem(msg % msgargs)

        self.makedirs()

        traditional_input_docs = list(self.doc.walk_input_docs())
        input_docs = traditional_input_docs + self.doc.additional_docs
        self.log_debug("input docs %s" % input_docs)

        for i, inpt in enumerate(input_docs):
            if not self.include_input_in_workspace(inpt):
                self.log_debug("not populating workspace with input '%s'" % inpt.key)
                continue

            data = inpt.output_data()

            filepath = data.name

            # Ensure parent dir exists.
            parent_dir = os.path.join(self.workspace(), os.path.dirname(filepath))
            if not parent_dir in already_created_dirs:
                try:
                    os.makedirs(parent_dir)
                    already_created_dirs.add(parent_dir)
                except OSError:
                    pass

            # Save contents of file to workspace
            self.log_debug("populating workspace with %s for %s" % (filepath, inpt.key))
            file_dest = os.path.join(self.workspace(), filepath)

            try:
                if i > len(traditional_input_docs)-1:
                    contents = inpt.setting('contents')
                    data.storage.write_data(contents, file_dest)
                else:
                    copy_or_link(data, file_dest, use_links=False)

            except Exception as e:
                self.log_debug("problem populating working dir with input %s" % data.key)
                self.log_debug(e)

            self._files_workspace_populated_with.add(filepath)

        self.input_data.output_to_file(self.work_input_filepath())
        rel_path_to_work_file = os.path.join(os.path.dirname(self.key), self.work_input_filename())
        self._files_workspace_populated_with.add(rel_path_to_work_file)

        self.custom_populate_workspace()

    def custom_populate_workspace(self):
        """
        Allow filters to run the standard populate_workspace, and also do extra
        things to workspace after populate_workspace runs. Filters can also
        just override populate_workspace.
        """
        pass

    def resolve_conflict(self, doc, conflict_docs):
        """
        Return true if the doc wins the conflict and should be written to the canonical name, false if not.
        """
        conflict_docs = [d for d in conflict_docs if not (('pyg' in d.key) or ('idio' in d.key))]
        conflict_docs.sort()
        if len(conflict_docs) == 0:
            return True
        else:
            return doc in conflict_docs and conflict_docs.index(doc) == 0

    def is_part_of_script_bundle(self):
        if hasattr(self.doc, 'parent'):
            return hasattr(self.doc.parent, 'script_storage')

    def script_storage(self):
        if not self.is_part_of_script_bundle():
            msg = "%s must be part of script bundle to access script storage"
            raise dexy.exceptions.UserFeedback(msg % self.key)
        return self.doc.parent.script_storage

class DexyFilter(Filter):
    """
    Filter which implements some default behaviors.
    """
    aliases = ['dexy']

    def process(self):
        if hasattr(self, "process_text"):
            output = self.process_text(str(self.input_data))
            self.output_data.set_data(output)
        else:
            self.output_data.copy_from_file(self.input_data.storage.data_file())

class AliasFilter(DexyFilter):
    """
    Filter to be used when an Alias is specified. Should not change input.
    """
    aliases = ['-']
    _settings = {
            'preserve-prior-data-class' : True
            }

    def calculate_canonical_name(self):
        return self.input_data.name

def filters_by_tag():
    """
    Returns a dict with tags as keys and lists of corresponding filter instances as values.
    """
    tags_filters = {}
    for filter_instance in Filter:
        if filter_instance.setting('nodoc'):
            continue

        for tag in filter_instance.setting('tags'):
            if not tag in tags_filters:
                tags_filters[tag] = []
            tags_filters[tag].append(filter_instance)

    return tags_filters

def filter_aliases_by_tag():
    tags_filters = filters_by_tag()
    tags = sorted(tags_filters.keys())
    return [(tag,
            [(filter_instance.alias, filter_instance.setting('help'))
                for filter_instance in sorted(tags_filters[tag], key=attrgetter('alias'))])
            for tag in tags]



import copy
import dexy.doc
import dexy.exceptions
import dexy.plugin
import posixpath

class AbstractSyntaxTree():
    def __init__(self, wrapper):
        self.wrapper = wrapper

        self.root_nodes_ordered = False

        self.lookup_table = {}
        self.tree = []

        # Lists of (directory, settings) tuples
        self.default_args_for_directory = []
        self.environment_for_directory = []

    def all_inputs(self):
        """
        Returns a set of all node keys identified as inputs of some other
        element.
        """
        all_inputs = set()
        for kwargs in list(self.lookup_table.values()):
            inputs = kwargs['inputs']
            all_inputs.update(inputs)
        return all_inputs

    def clean_tree(self):
        """
        Removes tasks which are already represented as inputs (tree should
        only contain root nodes).
        """
        treecopy = copy.deepcopy(self.tree)
        all_inputs = self.all_inputs()
        for task in treecopy:
            if task in all_inputs:
                self.tree.remove(task)

    def add_node(self, node_key, **kwargs):
        """
        Adds the node and its kwargs to the tree and lookup table
        """
        node_key = self.wrapper.standardize_key(node_key)

        if not node_key in self.tree:
            self.tree.append(node_key)

        if not node_key in self.lookup_table:
            self.lookup_table[node_key] = {}

        self.lookup_table[node_key].update(kwargs)

        if not 'inputs' in self.lookup_table[node_key]:
            self.lookup_table[node_key]['inputs'] = []

        self.clean_tree()
        return node_key

    def add_dependency(self, node_key, input_node_key):
        """
        Adds input_node_key to list of inputs for node_key (both nodes are
        also added to tree).
        """
        node_key = self.add_node(node_key)
        input_node_key = self.add_node(input_node_key)

        if not node_key == input_node_key:
            self.lookup_table[node_key]['inputs'].append(input_node_key)

        self.clean_tree()

    def args_for_node(self, node_key):
        """
        Returns the dict of kw args for a node
        """
        node_key = self.wrapper.standardize_key(node_key)
        args = copy.deepcopy(self.lookup_table[node_key])
        del args['inputs']
        return args

    def inputs_for_node(self, node_key):
        """
        Returns the list of inputs for a node
        """
        node_key = self.wrapper.standardize_key(node_key)
        return self.lookup_table[node_key]['inputs']

    def calculate_default_args_for_directory(self, path):
        dir_path = posixpath.dirname(posixpath.abspath(path))
        default_kwargs = {}

        for d, args in self.default_args_for_directory:
            if posixpath.abspath(d) in dir_path:
                default_kwargs.update(args)

        return default_kwargs

    def calculate_environment_for_directory(self, path):
        dir_path = posixpath.dirname(posixpath.abspath(path))
        env = {}

        for d, args in self.environment_for_directory:
            if posixpath.abspath(d) in dir_path:
                env.update(args)

        return env

    def walk(self):
        """
        Creates Node objects for all elements in tree. Returns a list of root
        nodes and a dict of all nodes referenced by qualified keys.
        """
        if self.wrapper.nodes:
            self.log_warn("nodes are not empty: %s" % ", ".join(self.wrapper.nodes))
        if self.wrapper.roots:
            self.log_warn("roots are not empty: %s" % ", ".join(self.wrapper.roots))

        def create_dexy_node(key, *inputs, **kwargs):
            """
            Stores already created nodes in nodes dict, if called more than
            once for the same key, returns already created node.
            """
            if not key in self.wrapper.nodes:
                alias, pattern = self.wrapper.qualify_key(key)
                node_environment = self.calculate_environment_for_directory(pattern)
                
                kwargs_with_defaults = self.calculate_default_args_for_directory(pattern)
                kwargs_with_defaults.update(kwargs)
                kwargs_with_defaults.update({'environment' : node_environment })

                self.wrapper.log.debug("creating node %s" % alias)
                node = dexy.node.Node.create_instance(
                        alias,
                        pattern,
                        self.wrapper,
                        inputs,
                        **kwargs_with_defaults)

                if node.inputs:
                    self.wrapper.log.debug("inputs are %s" % ", ".join(i.key for i in node.inputs))

                self.wrapper.nodes[key] = node

                for child in node.children:
                    self.wrapper.nodes[child.key_with_class()] = child

            return self.wrapper.nodes[key]

        def parse_item(key):
            inputs = self.inputs_for_node(key)
            kwargs = self.args_for_node(key)
            self.wrapper.log.debug("parsing item %s" % key)
            self.wrapper.log.debug("  inputs: %s" % ", ".join("%r" % inpt for inpt in inputs))
            self.wrapper.log.debug("  kwargs: %s" % ", ".join("%s: %r" % (k, v) for k, v in kwargs.items()))

            if kwargs.get('inactive') or kwargs.get('disabled'):
                return

            matches_target = self.wrapper.target and key.startswith(self.wrapper.target)
            if not kwargs.get('default', True) and not self.wrapper.full and not matches_target:
                return

            input_nodes = [parse_item(i) for i in inputs if i]
            input_nodes = [i for i in input_nodes if i]
    
            return create_dexy_node(key, *input_nodes, **kwargs)

        for node_key in self.tree:
            root_node = parse_item(node_key)
            if root_node:
                self.wrapper.roots.append(root_node)

class Parser(dexy.plugin.Plugin, metaclass=dexy.plugin.PluginMeta):
    """
    Parse various types of config file.
    """
    aliases = []
    _settings = {}

    def __init__(self, wrapper, ast):
        self.wrapper = wrapper
        self.ast = ast

    def file_exists(self, directory, filename):
        filepath = self.wrapper.join_dir(directory, filename)
        return self.wrapper.file_available(filepath)

    def parse(self, directory, input_text):
        pass



from dexy.exceptions import DeprecatedException
from dexy.exceptions import InternalDexyProblem
from dexy.exceptions import UserFeedback
from dexy.utils import file_exists
from dexy.utils import s
import chardet
import dexy.batch
import dexy.doc
import dexy.parser
import dexy.reporter
import dexy.utils
import logging
import logging.handlers
import os
import pickle
import posixpath
import shutil
import sys
import textwrap
import time
import traceback
import uuid

class Wrapper(object):
    """
    Class that manages run configuration and state and provides utilities such
    as logging and setting up/tearing down workspace on file system.
    """
    _required_dirs = ['artifacts_dir']

    state_transitions = (
            (None, 'new'),
            ('new', 'valid'),
            ('valid', 'valid'),
            ('valid', 'walked'),
            ('walked', 'checked'),
            ('checked', 'running'),
            ('running', 'error'),
            ('running', 'ran'),
            )

    def printmsg(self, msg):
        if self.silent:
            self.log.warn(msg)
        else:
            print(msg)

    def validate_state(self, state=None):
        """
        Checks that the instance is in the expected state, and validates
        attributes for the state.
        """
        if state:
            assert self.state == state
        else:
            state = self.state

        if state == 'new':
            assert not self.state_history
            assert self.project_root
        elif state == 'valid':
            pass
        elif state == 'walked':
            assert self.batch
            assert self.nodes
            assert self.roots
            assert self.filemap
            for node in self.nodes.values():
                assert node.state == 'new'
        elif state == 'checked':
            for node in self.nodes.values():
                assert node.state in ('uncached', 'consolidated', 'inactive',), node.state
        elif state == 'ran':
            for node in self.nodes.values():
                assert node.state in ('ran', 'consolidated', 'inactive'), node.state
        elif state == 'error':
            pass
        else:
            raise dexy.exceptions.InternalDexyProblem(state)

    def __init__(self, **kwargs):
        self.initialize_attribute_defaults()
        self.update_attributes_from_kwargs(kwargs)
        self.project_root = os.path.abspath(os.getcwd())
        self.project_root_ts = "%s%s" % (self.project_root, os.sep)
        self.state = None
        self.current_task = None
        self.lookup_nodes = {} # map of shortcuts/keys to all nodes which can match
        self.lookup_sections = {} # map of section names to nodes
        self.transition('new')

    def state_message(self):
        """
        A message to print at end of dexy run depending on the final wrapper state.
        """
        if self.state == 'error':
            return " WITH ERRORS"
        else:
            return ""

    def transition(self, new_state):
        dexy.utils.transition(self, new_state)

    def setup_for_valid(self):
        self.setup_log()

    def to_valid(self):
        if not self.dexy_dirs_exist():
            msg = "Should not attempt to enter 'valid' state unless dexy dirs exist."
            raise dexy.exceptions.InternalDexyProblem(msg)
        self.setup_for_valid()
        self.transition('valid')

    def walk(self):
        self.nodes = {}
        self.roots = []
        self.batch = dexy.batch.Batch(self)
        self.filemap = self.map_files()
        self.ast = self.parse_configs()
        self.ast.walk()

    def to_walked(self):
        self.walk()
        self.transition('walked')

    def check(self):
        # Clean and reset working dirs.
        self.reset_work_cache_dir()
        if not os.path.exists(self.this_cache_dir()):
            self.create_cache_dir_with_sub_dirs(self.this_cache_dir())

        # Load information about arguments from previous batch.
        self.load_node_argstrings()

        self.check_cache()
        self.consolidate_cache()

        # Save information about this batch's arguments for next time.
        self.save_node_argstrings()

    def check_cache(self):
        """
        Check whether all required files are already cached from a previous run
        """
        for node in self.roots:
            node.check_is_cached()

    def consolidate_cache(self):
        """
        Move all cache files from last/ cache to this/ cache
        """
        for node in self.roots:
            node.consolidate_cache_files()

        self.trash(self.last_cache_dir())

    def to_checked(self):
        self.check()
        self.transition('checked')

    # Cache dirs
    def this_cache_dir(self):
        return os.path.join(self.artifacts_dir, "this")

    def last_cache_dir(self):
        return os.path.join(self.artifacts_dir, "last")

    def work_cache_dir(self):
        return os.path.join(self.artifacts_dir, "work")

    def trash_dir(self):
        return os.path.join(self.project_root, ".trash")

    def create_cache_dir_with_sub_dirs(self, cache_dir):
        os.mkdir(cache_dir)
        hexes = ['0','1','2','3','4','5','6','7','8','9','a','b','c','d','e','f']
        for c in hexes:
            for d in hexes:
                os.mkdir(os.path.join(cache_dir, "%s%s" % (c,d)))

    def trash(self, d):
        """
        Move the passed file path (if it exists) into the .trash directory.
        """
        if not self.is_location_in_project_dir(d):
            msg = "trying to trash '%s', but this is not in project dir '%s"
            msgargs = (d, self.project_root)
            raise dexy.exceptions.InternalDexyProblem(msg % msgargs)

        trash_dir = self.trash_dir()

        try:
            os.mkdir(trash_dir)
        except OSError:
            pass

        move_to = os.path.join(trash_dir, str(uuid.uuid4()))

        try:
            shutil.move(d, move_to)
        except IOError:
            pass

    def empty_trash(self):
        for dirpath, dirnames, filenames in os.walk(self.trash_dir(), topdown=False):
            for f in filenames:
                filepath = os.path.join(dirpath, f)
                os.remove(filepath)

            try:
                os.rmdir(dirpath)
            except OSError as e:
                print("error removing dir")
                print(e)
                print(dirpath)
                print((os.listdir(dirpath)))
                print((os.stat(dirpath)))
                print((os.lstat(dirpath)))
                shutil.rmtree(dirpath)

    def reset_work_cache_dir(self):
        # remove work/ dir leftover from previous run (if any) and create a new
        # work/ dir for this run
        work_dir = self.work_cache_dir()
        self.trash(work_dir)
        self.create_cache_dir_with_sub_dirs(work_dir)

    def run(self):
        self.transition('running')

        self.batch.start_time = time.time()

        if self.target:
            matches = self.roots_matching_target()
        else:
            matches = self.roots

        try:
            for node in matches:
                for task in node:
                    task()

        except Exception as e:
            self.error = e
            self.transition('error')
            if self.debug:
                raise
            else:
                if self.current_task:
                    msg = "ERROR while running %s: %s\n" % (self.current_task.key, str(e))
                else:
                    msg = "ERROR: %s\n" % str(e)
                sys.stderr.write(msg)
                sys.stderr.write(traceback.format_exc())
                self.log.warn(msg)

        else:
            self.after_successful_run()

    def after_successful_run(self):
        self.transition('ran')
        self.batch.end_time = time.time()
        self.batch.save_to_file()
        shutil.move(self.this_cache_dir(), self.last_cache_dir())
        self.empty_trash()
        self.add_lookups()

    def add_lookups(self):
        for data in self.batch:
            data.add_to_lookup_sections()
            data.add_to_lookup_nodes()

    def bundle_docs(self):
        from dexy.node import BundleNode
        return [n for n in self.nodes.values() if isinstance(n, BundleNode)]

    def non_bundle_docs(self):
        from dexy.node import BundleNode
        return [n for n in self.nodes.values() if not isinstance(n, BundleNode)]

    def documents(self):
        from dexy.doc import Doc
        return [n for n in self.nodes.values() if isinstance(n, Doc)]

    def roots_matching_target(self):
        # First priority is to match any named bundles.
        matches = [n for n in self.bundle_docs() if n.key == self.target]
        if not matches:
            # Second priority is exact matches of any document key.
            matches = [n for n in self.non_bundle_docs() if n.key == self.target]
        if not matches:
            # Third priority is partial matches of any document key.
            matches = [n for n in self.nodes.values() if n.key.startswith(self.target)]

        if not matches:
            raise dexy.exceptions.UserFeedback("No matches found for '%s'" % self.target)

        self.log.debug("Documents matching target '%s' are: %s" % (self.target, ", ".join(m.key_with_class() for m in matches)))
        return matches

    def run_from_new(self):
        self.to_valid()
        self.to_walked()
        self.to_checked()
        if self.dry_run:
            self.printmsg("dry run only")
        else:
            self.run()

    def run_docs(self, *docs):
        self.to_valid()

        # do a custom walk() method
        self.roots = docs
        self.nodes = dict((node.key_with_class(), node) for node in self.roots)
        self.filemap = self.map_files()
        self.batch = dexy.batch.Batch(self)
        self.transition('walked')

        self.to_checked()
        self.run()

    # Attributes
    def initialize_attribute_defaults(self):
        """
        Applies the values in defaults dict to this wrapper instance.
        """
        for name, value in dexy.utils.defaults.items():
            setattr(self, name, value)

    def update_attributes_from_kwargs(self, kwargs):
        """
        Updates instance values from a dictionary of kwargs, checking that the
        attribute names are also present in defaults dict.
        """
        for key, value in kwargs.items():
            if not key in dexy.utils.defaults:
                msg = "invalid kwarg '%s' being passed to wrapper, not defined in defaults dict" 
                raise InternalDexyProblem(msg % key)
            setattr(self, key, value)

    def node_argstrings_filename(self):
        return os.path.join(self.artifacts_dir, 'batch.args.pickle')

    def save_node_argstrings(self):
        """
        Save string representation of node args to check if they have changed.
        """
        arg_info = {}

        for node in self.nodes.values():
            arg_info[node.key_with_class()] = node.sorted_arg_string()

        with open(self.node_argstrings_filename(), 'wb') as f:
            pickle.dump(arg_info, f)

    def load_node_argstrings(self):
        """
        Load saved node arg strings into a hash so nodes can check if their
        args have changed.
        """
        try:
            with open(self.node_argstrings_filename(), 'rb') as f:
                self.saved_args = pickle.load(f)
        except IOError:
            self.saved_args = {}

    # Dexy Dirs
    def iter_dexy_dirs(self):
        """
        Iterate over the required dirs (e.g. artifacts, logs)
        """
        for d in self.__class__._required_dirs:
            dirpath = self.__dict__[d]
            safety_filepath = os.path.join(dirpath, self.safety_filename)
            try:
                stat = os.stat(dirpath)
            except OSError:
                stat = None

            if stat:
                if not file_exists(safety_filepath):
                    msg = s("""You need to manually delete the '%s' directory
                    and then run 'dexy setup' to create new directories. This
                    should just be a once-off issue due to a change in dexy to
                    prevent accidentally deleting directories which dexy does
                    not create.
                    """) % dirpath
                    raise UserFeedback(msg)

            yield (dirpath, safety_filepath, stat)

    def dexy_dirs_exist(self):
        """
        Returns a boolean indicating whether dexy dirs exist.
        """
        return all(file_exists(d[0]) for d in self.iter_dexy_dirs())

    def assert_dexy_dirs_exist(self):
        """
        Raise a UserFeedback error if user has tried to run dexy without
        setting up necessary directories first.
        """
        self.detect_dot_dexy_files()
        self.update_cache_directory()

        if not self.dexy_dirs_exist():
            msg = "You need to run 'dexy setup' in this directory first."
            raise UserFeedback(msg)

        self.deprecate_logs_directory()

    def create_dexy_dirs(self):
        """
        Creates the directories needed for dexy to run. Does not complain if
        directories are already present.
        """
        for dirpath, safety_filepath, dirstat in self.iter_dexy_dirs():
            if not dirstat:
                os.mkdir(dirpath)
                with open(safety_filepath, 'w') as f:
                    f.write("This directory was created by dexy.")

    def detect_dot_dexy_files(self):
        if os.path.exists(".dexy") and os.path.isfile(".dexy"):
            msg = """\
            You have a file named '.dexy' in your project, this format is no longer supported.
            See http://dexy.it/guide/getting-started.html for updated tutorials.
            Please rename or remove this file to proceed."""
            raise dexy.exceptions.UserFeedback(textwrap.dedent(msg))

    def update_cache_directory(self):
        """
        Move .cache directories to be .dexy directories.
        """
        old_cache_dir = ".cache"
        safety_file = os.path.join(old_cache_dir, self.safety_filename)

        if self.artifacts_dir == old_cache_dir:
            msg = """\
            You may have a dexy.conf file present in which you specify
            'artifactsdir: .cache'
            
            Dexy's defaults have changed and the .cache directory is being
            renamed to .dexy

            The easiest way to never see this message again is to remove this
            line from your dexy.conf file (You should remove any entries in
            dexy.conf which you don't specifically want to customize.)
            
            If you really want the artifacts directory to be at .cache then you
            can leave your config in place and this message will disappear in a
            future dexy version.
            """
            print((textwrap.dedent(msg)))

        elif os.path.exists(old_cache_dir) and os.path.isdir(old_cache_dir) and os.path.exists(safety_file):
            if os.path.exists(self.artifacts_dir):
                if os.path.isdir(self.artifacts_dir):
                    msg = "You have a dexy '%s' directory and a '%s' directory. Please remove '%s' or at least '%s'."
                    msgargs = (old_cache_dir, self.artifacts_dir, old_cache_dir, safety_file)
                    raise dexy.exceptions.UserFeedback(msg % msgargs)
                else:
                    msg = "'%s' is not a dir!" % (self.artifacts_dir,)
                    raise dexy.exceptions.InternalDexyProblem(msg)

            print(("Moving directory '%s' to new location '%s'" % (old_cache_dir, self.artifacts_dir)))
            shutil.move(old_cache_dir, self.artifacts_dir)

    def deprecate_logs_directory(self):
        log_dir = 'logs'
        if self.log_dir != self.artifacts_dir:
            # user has set a custom log dir
            log_dir = self.log_dir

        safety_file = os.path.join(log_dir, self.safety_filename)
        deprecation_notice_file = os.path.join(log_dir, "WHERE-ARE-THE-LOGS.txt")

        if os.path.exists(log_dir) and os.path.exists(safety_file) and not os.path.exists(deprecation_notice_file):
            deprecation_notice = """\
            Dexy no longer has a separate '{log_dir}' directory for the log file.
            The logfile can now be found at: {log_path}\n""".format(
                    log_path = self.log_path(),
                    log_dir = log_dir)
            
            deprecation_notice = "\n".join(l.strip() for l in deprecation_notice.splitlines())
            self.printmsg("Deprecating %s/ directory" % log_dir)
            self.printmsg(deprecation_notice)
            self.printmsg("You can remove the %s/ directory" % log_dir)

            with open(deprecation_notice_file, 'w') as f:
                f.write(deprecation_notice + "\n\nYou can remove this directory.\n")
            self.trash(os.path.join(log_dir, "dexy.log"))

    def remove_dexy_dirs(self):
        for dirpath, safety_filepath, dirstat in self.iter_dexy_dirs():
            if dirstat:
                self.trash(dirpath)
        self.empty_trash()

    def remove_reports_dirs(self, reports=True, keep_empty_dir=False):
        if reports:
            if isinstance(reports, bool):
                # return an iterator over all reporters
                reports=dexy.reporter.Reporter

            for report in reports:
                report.remove_reports_dir(self, keep_empty_dir)

    # Logging
    def log_path(self):
        """
        Returns path to logfile.
        """
        return os.path.join(self.artifacts_dir, self.log_file)

    def setup_log(self):
        """
        Creates a logger and assigns it to 'log' attribute of wrapper.
        """
        formatter = logging.Formatter(self.log_format)
        log_level = dexy.utils.logging_log_level(self.log_level)

        handler = logging.handlers.RotatingFileHandler(
                self.log_path(),
                encoding="utf-8")

        handler.setFormatter(formatter)

        self.log = logging.getLogger('dexy')
        self.log.setLevel(log_level)
        self.log.addHandler(handler)
        self.log.info("starting logging for dexy")
        self.log_handler = handler

    def flush_logs(self):
        self.log_handler.flush()

    # Project files
    def exclude_dirs(self):
        """
        Returns list of directory names which should be excluded from dexy processing.
        """
        exclude_str = self.exclude
        if self.exclude_also:
            exclude_str += ",%s" % self.exclude_also

        exclude = [d.strip() for d in exclude_str.split(",")]
        exclude += self.reports_dirs()

        for d in self.iter_dexy_dirs():
            exclude += [d[0], "%s-old" % d[0]]

        return exclude

    def reports_dirs(self):
        """
        Returns list of directories which are written to by reporters.
        """
        dirs_and_nones = [i.setting('dir') for i in dexy.reporter.Reporter]
        return [d for d in dirs_and_nones if d]

    def map_files(self):
        """
        Generates a map of files present in the project directory.
        """
        exclude = self.exclude_dirs()
        filemap = {}

        for dirpath, dirnames, filenames in os.walk('.', followlinks=True):
            for x in exclude:
                if x in dirnames and not x in self.include:
                    dirnames.remove(x)

            if '.nodexy' in filenames:
                dirnames[:] = []
            elif 'pip-delete-this-directory.txt' in filenames:
                msg = s("""pip left an old build/ file lying around,
                please remove this before running dexy""")
                raise UserFeedback(msg)
            else:
                for filename in filenames:
                    filepath = posixpath.normpath(posixpath.join(dirpath, filename))
                    filemap[filepath] = {}
                    filemap[filepath]['stat'] = os.stat(os.path.join(dirpath, filename))
                    filemap[filepath]['ospath'] = os.path.normpath(os.path.join(dirpath, filename))
                    filemap[filepath]['dir'] = os.path.normpath(dirpath)

        return filemap

    def file_available(self, filepath):
        """
        Does the file exist and is it available to dexy?
        """
        return filepath in self.filemap

    # Running Dexy
    def add_node(self, node):
        """
        Add new nodes which are not children of other nodes.
        """
        key = node.key_with_class()
        self.nodes[key] = node

    def add_data_to_lookup_nodes(self, key, data):
        if not key in self.lookup_nodes:
            self.lookup_nodes[key] = []
        if not data in self.lookup_nodes[key]:
            self.lookup_nodes[key].append(data)

    def add_data_to_lookup_sections(self, key, data):
        if not key in self.lookup_sections:
            self.lookup_sections[key] = []
        if not data in self.lookup_sections[key]:
            self.lookup_sections[key].append(data)

    def qualify_key(self, key):
        """
        A full node key is of the form alias:pattern where alias indicates
        the type of node to be created. This method determines the alias if it
        is not specified explicitly, and returns the alias, pattern tuple.
        """
        if not key:
            msg = "trying to call qualify_key with key of '%s'!"
            raise DeprecatedException(msg % key)

        if ":" in key:
            # split qualified key into alias & pattern
            alias, pattern = key.split(":")
        else:
            # this is an unqualified key, figure out its alias
            pattern = key

            # Allow '.ext' instead of '*.ext', shorter + easier for YAML
            if pattern.startswith(".") and not pattern.startswith("./"):
                if not self.file_available(pattern):
                    pattern = "*%s" % pattern

            filepath = pattern.split("|")[0]
            if self.file_available(filepath):
                alias = 'doc'
            elif (not "." in pattern) and (not "|" in pattern):
                alias = 'bundle'
            elif "*" in pattern:
                alias = 'pattern'
            else:
                alias = 'doc'

        return alias, pattern

    def standardize_alias(self, alias):
        """
        Nodes can have multiple aliases, standardize on first one in list.
        """
        # TODO should we just make it so nodes only have 1 alias?
        node_class, _ = dexy.node.Node.plugins[alias]
        return node_class.aliases[0]

    def standardize_key(self, key):
        """
        Standardizes the key by making the alias explicit and standardized, so
        we don't create 2 entires in the AST for what turns out to be the same
        node/task.
        """
        alias, pattern = self.qualify_key(key)
        alias = self.standardize_alias(alias)
        return "%s:%s" % (alias, pattern)

    def join_dir(self, directory, key):
        if directory == ".":
            return key
        else:
            starts_with_dot = key.startswith(".") and not key.startswith("./")
            if starts_with_dot:
                path_to_key = os.path.join(directory, key)
                if not self.file_available(path_to_key):
                    key = "*%s" % key
            return posixpath.join(directory, key)

    def explicit_config_files(self):
        return [c.strip() for c in self.configs.split()]

    def is_explicit_config(self, filepath):
        return filepath in self.explicit_config_files()

    def parse_configs(self):
        """
        Look for document config files in current working tree and load them.
        Return an Abstract Syntax Tree with information about nodes to be
        processed.
        """
        ast = dexy.parser.AbstractSyntaxTree(self)

        processed_configs = []

        for alias in self.parsers.split():
            parser = dexy.parser.Parser.create_instance(alias, self, ast)

            for filepath, fileinfo in self.filemap.items():
                if fileinfo['dir'] == '.' or self.recurse or self.is_explicit_config(filepath):
                    if os.path.split(filepath)[1] == alias:
                        self.log.info("using config file '%s'" % filepath)

                        config_file = fileinfo['ospath']
                        dirname = fileinfo['dir']

                        with open(config_file, "r") as f:
                            config_text = f.read()

                        try:
                            processed_configs.append(filepath)
                            parser.parse(dirname, config_text)
                        except UserFeedback:
                            sys.stderr.write("Problem occurred while parsing %s\n" % config_file)
                            raise

        if len(processed_configs) == 0:
            msg = "didn't find any document config files (like %s)"
            self.printmsg(msg % self.parsers)

        return ast

    def report(self):
        if self.reports:
            self.log.debug("generating user-specified reports '%s'" % self.reports)
            reporters = []
            for alias in self.reports.split():
                reporter = dexy.reporter.Reporter.create_instance(alias)
                reporters.append(reporter)
        else:
            msg = "no reports specified, running default reports"
            self.log.debug(msg)
            reporters = [i for i in dexy.reporter.Reporter if i.setting('default')]

        for reporter in reporters:
            if self.state in reporter.setting('run-for-wrapper-states'):
                self.log.debug("running reporter %s" % reporter.aliases[0])
                reporter.run(self)

    def is_location_in_project_dir(self, filepath):
        return self.writeanywhere or (self.project_root_ts in os.path.abspath(filepath))

    def decode_encoded(self, text):
        if self.encoding == 'chardet':
            encoding = chardet.detect(text)['encoding']
            if not encoding:
                return text.decode("utf-8")
            else:
                return text.decode(encoding)
        else:
            return text.decode(self.encoding)




from setuptools import setup, find_packages
from dexy.version import DEXY_VERSION

import platform
is_windows = platform.system() == 'Windows'

if is_windows:
    os_specific_requires = []
else:
    os_specific_requires = ['pexpect']

setup(
        author='Ana Nelson',
        author_email='ana@ananelson.com',
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Environment :: Console",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Financial and Insurance Industry",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: MIT License",
            "Topic :: Documentation",
            "Topic :: Software Development :: Build Tools",
            "Topic :: Software Development :: Code Generators",
            "Topic :: Software Development :: Documentation",
            "Topic :: Text Processing",
            "Topic :: Text Processing :: Markup :: HTML",
            "Topic :: Text Processing :: Markup :: LaTeX"
            ],
        description='Document Automation',
        ### "entry-points"
        entry_points = {
            'console_scripts' : [
                'dexy = dexy.commands:run'
                ],
            'pygments.lexers' : [
                'rst+django = dexy.filters.utils:RstDjangoLexer'
                ]
            },
        ### @end
        include_package_data = True,
        install_requires = os_specific_requires + [
            # for internal dexy use or used in many common plugins
            'BeautifulSoup4',
            'PyYAML',
            'cashew>=0.4.1',
            'chardet',
            'inflection>=0.2.0',
            'jinja2',
            'ply>=3.4',
            'pygments',
            'python3-modargs',
            'requests>=0.10.6',
            # for convenience of running additional filters
            'Markdown',
            'docutils'
            ],
        name='dexy',
        packages=find_packages(),
        url='http://dexy.it',
        version=DEXY_VERSION
    )



from dexy.filters.pexp import PexpectReplFilter
from dexy.filters.process import SubprocessCompileFilter
from dexy.filters.process import SubprocessStdoutFilter
import os
import platform

class Scala(SubprocessCompileFilter):
    """
    Compiles and runs .scala files.
    """
    aliases = ['scala']
    _settings = {
            'executable' : 'scalac',
            'tags' : ['code', 'scala', 'compiled', 'jvm'],
            'compiler-command-string' : "%(prog)s %(compiler_args)s %(script_file)s",
            'compiled-extension' : '',
            'input-extensions' : ['.scala', '.txt'],
            'output-extensions' : ['.txt']
            }

    def run_command_string(self):
        args = self.default_command_string_args()
        args['compiled_filename'] = self.compiled_filename()
        return "scala %(compiled_filename)s %(args)s" % args

class CompileScala(Scala):
    """
    Compiles .scala code to .class files.
    """
    aliases = ['scalac']
    _settings = {
            'tags' : ['code', 'scala', 'compiled', 'jvm'],
            'output-extensions' : ['.class']
            }

    def process(self):
        # Compile the code
        command = self.compile_command_string()
        proc, stdout = self.run_command(command, self.setup_env())
        self.handle_subprocess_proc_return(command, proc.returncode, stdout)
        self.copy_canonical_file()

class JythonFilter(SubprocessStdoutFilter):
    """
    jython
    """
    aliases = ['jython']
    _settings = {
            'executable' : 'jython',
            'tags' : ['code', 'jython', 'jvm'],
            'input-extensions' : [".py", ".txt"],
            'output-extensions' : [".txt"],
            'version-command' : "jython --version"
            }

    def is_active(klass):
        if platform.system() in ('Linux', 'Windows'):
            return True
        elif platform.system() in ('Darwin'):
            if hasattr(klass, 'log'):
                klass.log.warn("The jython dexy filter should not be run on MacOS due to a serious bug. This filter is being disabled.")
            return False
        else:
            if hasattr(klass, 'log'):
                klass.log.warn("""Can't detect your system. If you see this message please report this to the dexy project maintainer, your platform.system() value is '%s'. The jython dexy filter should not be run on MacOS due to a serious bug.""" % platform.system())
            return True

class JythonInteractiveFilter(PexpectReplFilter):
    """
    jython in REPL
    """
    aliases = ['jythoni']
    _settings = {
            'check-return-code' : False,
            'tags' : ['code', 'jython', 'jvm', 'repl'],
            'executable' : 'jython -i',
            'initial-timeout' : 30,
            'input-extensions' : [".py", ".txt"],
            'output-extensions' : [".pycon"],
            'version-command' : "jython --version"
            }

    def is_active(klass):
        if platform.system() in ('Linux', 'Windows'):
            return True
        elif platform.system() in ('Darwin'):
            print("The jythoni dexy filter should not be run on MacOS due to a serious bug. This filter is being disabled.")
            return False
        else:
            print("""Can't detect your system. If you see this message please report this to the dexy project maintainer, your platform.system() value is '%s'. The jythoni dexy filter should not be run on MacOS due to a serious bug.""" % platform.system())
            return True

class JavaFilter(SubprocessCompileFilter):
    """
    Compiles java code and runs main method.
    """
    aliases = ['java']
    _settings = {
            'check-return-code' : True,
            'classpath' : ("Custom entries in classpath.", []),
            'tags' : ['code', 'java', 'jvm', 'compiled'],
            'executable' : 'javac',
            'input-extensions' : ['.java'],
            'output-extensions' : ['.txt'],
            'main' : ("Main method.", None),
            'version-command' : 'java -version',
            'compiled-extension' : ".class",
            'compiler-command-string' : "%(prog)s %(compiler_args)s %(classpath)s %(script_file)s"
            }

    def setup_cp(self):
        """
        Makes sure the current working directory is on the classpath, also adds
        any specified CLASSPATH elements. Assumes that CLASSPATH elements are either
        absolute paths, or paths relative to the artifacts directory. Also, if
        an input has been passed through the javac filter, its directory is
        added to the classpath.
        """
        self.log_debug("in setup_cp for %s" % self.key)

        classpath_elements = []

        working_dir = os.path.join(self.workspace(), self.output_data.parent_dir())
        abs_working_dir = os.path.abspath(working_dir)
        self.log_debug("Adding working dir %s to classpath" % abs_working_dir)
        classpath_elements.append(abs_working_dir)

        for doc in self.doc.walk_input_docs():
            if (doc.output_data().ext == ".class") and ("javac" in doc.key):
                classpath_elements.append(doc.output_data().parent_dir())

        for item in self.setting('classpath'):
            for x in item.split(":"):
                classpath_elements.append(x)

        env = self.setup_env()
        if env and 'CLASSPATH' in env:
            for x in env['CLASSPATH'].split(":"):
                classpath_elements.append(x)

        cp = ":".join(classpath_elements)
        self.log_debug("Classpath %s" % cp)
        return cp

    def compile_command_string(self):
        args = self.default_command_string_args()
        args['compiler_args'] = self.setting('compiler-args')

        # classpath
        cp = self.setup_cp()
        if len(cp) == 0:
            args['classpath'] = ''
        else:
            args['classpath'] = "-classpath %s" % cp

        return self.setting('compiler-command-string') % args

    def run_command_string(self):
        args = self.default_command_string_args()
        args['main_method'] = self.setup_main_method()

        # classpath
        cp = self.setup_cp()
        if len(cp) == 0:
            args['classpath'] = ''
        else:
            args['classpath'] = "-cp %s" % cp

        return "java %(args)s %(classpath)s %(main_method)s" % args

    def setup_main_method(self):
        basename = os.path.basename(self.input_data.name)
        default_main = os.path.splitext(basename)[0]
        if self.setting('main'):
            return self.setting('main')
        else:
            return default_main

class JavacFilter(JavaFilter):
    """
    Compiles java code and returns the .class object
    """
    aliases = ['javac']

    _settings = {
            'executable' : 'javac',
            'input-extensions' : ['.java'],
            'output-extensions' : ['.class'],
            'tags' : ['code', 'java', 'jvm', 'compiled'],
            'version-command' : 'java -version'
            }

    def process(self):
        # Compile the code
        command = self.compile_command_string()
        proc, stdout = self.run_command(command, self.setup_env())
        self.handle_subprocess_proc_return(command, proc.returncode, stdout)
        self.copy_canonical_file()



from tests.utils import assert_in_output

def test_easyhtml_filter():
    some_html = "<p>This is some HTML</p>"
    assert_in_output("easyhtml", some_html, some_html, ".html")
    assert_in_output("easyhtml", some_html, "<html>", ".html")



"""
DexyFilters written to be examples of how to write filters.
"""
from dexy.doc import Doc
from dexy.filter import DexyFilter
import os

class Example(DexyFilter):
    """
    Examples of how to write filters.
    """
    aliases = []
    NODOC = True

class KeyValueExample(Example):
    """
    Example of storing key value data.
    """
    aliases = ['keyvalueexample']

    _settings = {
            'data-type' : 'keyvalue',
            'output-extensions' : ['.sqlite3', '.json']
            }

    def process(self):
        assert self.output_data.state == 'ready'
        self.output_data.append("foo", "bar")
        self.output_data.save()

class AccessOtherDocuments(Example):
    """
    Example of accessing other documents.
    """
    aliases = ["others"]

    def process_text(self, input_text):
        info = []
        info.append("Here is a list of previous docs in this tree (not including %s)." % self.key)

        ### @export "access-other-docs-iterate"
        for doc in self.doc.walk_input_docs():
            assert isinstance(doc, Doc)

            ### @export "access-other-docs-lens"
            n_children = len(doc.children)
            n_inputs = len(doc.inputs)

            ### @export "access-other-docs-output-length"
            if doc.output_data().has_data():
                length = len(doc.output_data().data())
            else:
                length = len(doc.output_data().ordered_dict())

            ### @export "access-other-docs-finish"
            info.append("%s (%s children, %s inputs, length %s)" % (doc.key, n_children, n_inputs, length))
        s = "%s        " % os.linesep
        return s.join(info)
        ### @end

class AddNewDocument(Example):
    """
    A filter which adds an extra document to the tree.
    """
    aliases = ['newdoc']

    def process_text(self, input_text):
        self.add_doc("newfile.txt|processtext", "newfile")
        return "we added a new file"

class ConvertDict(Example):
    """
    Returns an ordered dict with a single element.
    """
    aliases = ['dict']

    def process(self, input_text):
        self.output_data['1'] = str(self.input_data)
        self.output_data.save()

class ExampleProcessTextMethod(Example):
    """
    Uses process_text method
    """
    aliases = ['processtext']

    def process_text(self, input_text):
        return "Dexy processed the text '%s'" % input_text

class ExampleProcessMethod(Example):
    """
    Calls `set_data` method to store output.
    """
    aliases = ['process']

    def process(self):
        output = "Dexy processed the text '%s'" % self.input_data
        self.output_data.set_data(output)

class ExampleProcessMethodManualWrite(Example):
    """
    Writes output directly to output file.
    """
    aliases = ['processmanual']

    def process(self):
        input_data = self.input_data
        output = "Dexy processed the text '%s'" % input_data
        with open(self.output_filepath(), "w") as f:
            f.write(output)

class ExampleProcessWithDictMethod(Example):
    """
    Stores sectional data using `process` method.
    """
    aliases = ['processwithdict']
    _settings = {
            'data-type' : 'sectioned'
            }

    def process(self):
        self.output_data['1'] = "Dexy processed the text '%s'" % self.input_data
        self.output_data.save()

class AbcExtension(Example):
    """
    Only outputs extension .abc
    """
    aliases = ['outputabc']
    _settings = {
            'output-extensions' : ['.abc']
            }

    def process_text(self, input_text):
        return "Dexy processed the text '%s'" % input_text

class ExampleFilterArgs(Example):
    """
    Prints out the args it receives.
    """
    aliases = ['filterargs']

    _settings = {
            "abc" : ("The abc setting.", None),
            "foo" : ("The foo setting.", None),
            }

    def process_text(self, input_text):
        # Filter Settings
        result = ["Here are the filter settings:"]
        for k in sorted(self.setting_values()):
            v = self.setting_values()[k]
            result.append("  %s: %s" % (k, v))

        # Doc args
        result.append("Here are the document args:")
        for k in sorted(self.doc.args):
            v = self.doc.args[k]
            result.append("  %s: %s" % (k, v))

        return os.linesep.join(result)



from dexy.doc import Doc
from tests.utils import TEST_DATA_DIR
from tests.utils import assert_output
from tests.utils import runfilter
from tests.utils import wrap
from nose.exc import SkipTest
import os
import shutil

def test_phantomjs_render_filter():
    with runfilter("phrender", "<p>hello</p>") as doc:
        assert doc.output_data().is_cached()

def test_phantomjs_stdout_filter():
    assert_output('phantomjs', PHANTOM_JS, "Hello, world!\n")

def test_casperjs_svg2pdf_filter():
    raise SkipTest() # TODO fix this - if casper is missing should raise error before reach assertions
    # TODO find smaller file - make test go faster?
    with wrap() as wrapper:
        orig = os.path.join(TEST_DATA_DIR, 'butterfly.svg')
        shutil.copyfile(orig, 'butterfly.svg')

        from dexy.wrapper import Wrapper
        wrapper = Wrapper()

        node = Doc("butterfly.svg|svg2pdf", wrapper)

        wrapper.run_docs(node)

        assert node.output_data().is_cached()
        assert node.output_data().filesize() > 1000

def test_casperjs_stdout_filter():
    with wrap() as wrapper:
        node = Doc("example.js|casperjs",
                wrapper,
                [],
                contents=CASPER_JS,
                casperjs={"add-new-files" : True }
                )

        wrapper.run_docs(node)

        try:
            assert 'doc:google.pdf' in wrapper.nodes
            assert 'doc:cookies.txt' in wrapper.nodes
        except AssertionError:
            pass

PHANTOM_JS = """
console.log('Hello, world!');
phantom.exit();
"""

CASPER_JS = """
var links = [];
var casper = require('casper').create();

casper.start('http://google.com/', function() {
    this.capture('google.pdf');
});

casper.run();
"""



from dexy.doc import Doc
from dexy.filters.templating import TemplateFilter
from dexy.filters.templating_plugins import TemplatePlugin
from tests.utils import run_templating_plugin as run
from tests.utils import wrap
import dexy.filters.templating_plugins as plugin
import inspect
import os

def test_d_object():
    with wrap() as wrapper:
        node = Doc("template.txt|jinja",
                wrapper,
                [Doc("input.txt",
                    wrapper,
                    [],
                    contents = "I am the input.")
                    ],
                contents = "The input is '{{ d['input.txt'] }}'")

        wrapper.run_docs(node)
        env = node.filters[0].run_plugins()
        d = env['d'][1]
        assert d.__class__.__name__ == 'D'
        assert len(list(d.items())) == 1

def test_base():
    run(TemplatePlugin)

def test_ppjson():
    with run(plugin.PrettyPrintJson) as env:
        assert 'ppjson' in env
        assert hasattr(env['ppjson'][1], '__call__')

def test_python_datetime():
    with run(plugin.PythonDatetime) as env:
        assert inspect.ismodule(env['cal'][1])

def test_dexy_version():
    with run(plugin.DexyVersion) as env:
        assert env['DEXY_VERSION'][1]

def test_simple_json():
    with run(plugin.SimpleJson) as env:
        assert inspect.ismodule(env['json'][1])

def test_python_builtins():
    with run(plugin.PythonBuiltins) as env:
        assert 'hasattr' in env

def test_pygments():
    with run(plugin.PygmentsStylesheet) as env:
        assert 'pastie.tex' in list(env['pygments'][1].keys())
        assert 'pastie.css' in list(env['pygments'][1].keys())
        assert 'pastie.html' in list(env['pygments'][1].keys())

class TestSubdirectory(TemplateFilter):
    """
    test subdir
    """
    aliases = ['testsubdir']
    _settings = { 'plugins' : ['subdirectories'] }

def test_subdirectories():
    with wrap() as wrapper:
        os.makedirs("s1")
        os.makedirs("s2")

        node = Doc("file.txt|testsubdir",
                wrapper,
                [],
                contents="hello"
                )

        wrapper.run_docs(node)

        env = node.filters[-1].run_plugins()
        assert 's1' in env['subdirectories'][1]
        assert 's2' in env['subdirectories'][1]

class TestVariables(TemplateFilter):
    """
    test variables
    """
    aliases = ['testvars']
    _settings = { 'plugins' : ['variables'] }

def test_variables():
    with wrap() as wrapper:
        node = Doc("hello.txt|testvars",
                wrapper,
                [],
                contents = "hello",
                testvars = { "variables" : {"foo" : "bar", "x" : 123.4 } }
                )

        wrapper.run_docs(node)

        env = node.filters[-1].run_plugins()
        assert env['foo'][1] == 'bar'
        assert env['x'][1] == 123.4

class TestGlobals(TemplateFilter):
    """
    test globals
    """
    aliases = ['testglobals']
    _settings = { 'plugins' : ['globals'] }

def test_globals():
    with wrap() as wrapper:
        wrapper.globals = "foo=bar"
        node = Doc("hello.txt|testglobals",
                wrapper,
                [],
                contents = "hello"
                )

        wrapper.run_docs(node)
        env = node.filters[-1].run_plugins()
        assert env['foo'][1] == 'bar'



from dexy.commands.utils import init_wrapper
from dexy.commands.utils import template_text
from dexy.utils import getdoc
import dexy.templates
import os
import sys
from dexy.utils import file_exists

DEFAULT_TEMPLATE = 'dexy:default'
def gen_command(
        plugins='', # extra python packages to load so plugins will register with dexy
        d=None,  # The directory to place generated files in, must not exist.
        t=False, # Shorter alternative to --template.
        template=DEFAULT_TEMPLATE, # The alias of the template to use.
        **kwargs # Additional kwargs passed to template's run() method.
        ):
    """
    Generate a new dexy project in the specified directory, using the template.
    """
    wrapper = init_wrapper(locals())

    if t and (template == DEFAULT_TEMPLATE):
        template = t
    elif t and template != DEFAULT_TEMPLATE:
        raise dexy.exceptions.UserFeedback("Only specify one of --t or --template, not both.")

    if not template in dexy.template.Template.plugins:
        print("Can't find a template named '%s'. Run 'dexy templates' for a list of templates." % template)
        sys.exit(1)

    template_instance = dexy.template.Template.create_instance(template)
    template_instance.generate(d, **kwargs)

    # We run dexy setup. This will respect any dexy.conf file in the template
    # but passing command line options for 'setup' to 'gen' currently not supported.
    os.chdir(d)
    wrapper.create_dexy_dirs()
    print("Success! Your new dexy project has been created in directory '%s'" % d)
    if file_exists("README"):
        print("\n--------------------------------------------------")
        with open("README", "r") as f:
            print(f.read())
        print("\n--------------------------------------------------")
        print("\nThis information is in the 'README' file for future reference.")

def template_command(
        alias=None
        ):
    print(template_text(alias))

def templates_command(
        plugins='', # extra python packages to load so plugins will register with dexy
        simple=False, # Only print template names, without docstring or headers.
        validate=False, # Intended for developer use only, validate templates (runs and checks each template).
        key=False # Only print information which matches this search key.
        ):
    """
    List templates that can be used to generate new projects.
    """
    init_wrapper(locals())

    if not simple:
        FMT = "%-40s %s"
        print(FMT % ("Alias", "Info"))

    for i, template in enumerate(dexy.template.Template):
        if key:
            if not key in template.alias:
                continue

        if template.setting('nodoc'):
            continue

        if simple:
            print(template.alias)
        else:
            first_line_help = template.setting('help').splitlines()[0].strip()
            print(FMT % (template.alias, first_line_help), end=' ')
            if validate:
                print(" validating...", end=' ')
                print(template.validate() and "OK" or "ERROR")
            else:
                print('')
    
    if i < 5:
        print("Run '[sudo] pip install dexy-templates' to install some more templates.")

    if not simple:
        print("Run 'dexy help -on gen' for help on generating projects from templates.")



from dexy.exceptions import UserFeedback
from dexy.filters.process import SubprocessFilter
import os
import re
import subprocess
import json

try:
    import IPython.nbformat.current
    AVAILABLE = True
except ImportError:
    AVAILABLE = False

class IPythonCasper(SubprocessFilter):
    """
    Launch IPython notebook and run a casperjs script against the server.
    """
    aliases = ['ipynbcasper']

    _settings = {
            'add-new-files' : True,
            'added-in-version' : "0.9.9.6",
            'examples' : ['ipynbcasper'],
            'args' : '--web-security=false --ignore-ssl-errors=true',
            'cell-timeout' : ("Timeout (in microseconds) for running individual notebook cells.", 5000),
            'command-string' : "%(prog)s %(test)s %(args)s %(script)s",
            'test' : ("Whether to run casperjs as 'test' mode.", False),
            'executable' : 'casperjs',
            'height' : ('Height of page to capture.', 5000),
            'image-ext' : ("File extension of images to capture.", ".png"),
            'input-extensions' : ['.ipynb'],
            'ipython-args' : ("Additional args to pass to ipython notebook command (list of string args).", None),
            'ipython-dir' : ("Directory in which to launch ipython, defaults to temp working dir.", None),
            'output-extensions' : ['.json', '.txt'],
            'script' : ("Canonical name of input document to use as casper script.", "full.js"),
            'timeout' : ("Timeout for the casperjs subprocess.", 10000),
            'version-command' : 'casperjs --version',
            'width' : ('Width of page to capture.', 800),
            }

    def is_active(self):
        return AVAILABLE

    def command_string_args(self):
        args = self.default_command_string_args()
        if self.setting('test'):
            args['test'] = 'test'
        else:
            args['test'] = ''
        return args

    def configure_casper_script(self, wd, port, cellmetas):
        scriptfile = os.path.join(wd, self.setting('script'))
        cellmetafile = os.path.join(wd, "%s-cellmetas.js" % self.input_data.baserootname())

        default_scripts_dir = os.path.join(os.path.dirname(__file__), "ipynbcasper")

        if not os.path.exists(scriptfile):
            # look for a matching default script
            script_setting = self.setting('script')

            filepath = os.path.join(default_scripts_dir, script_setting)
            if os.path.exists(filepath):
                with open(filepath, "r") as f:
                    js = f.read()
            else:
                default_scripts = os.listdir(default_scripts_dir)
                args = (self.setting('script'), ", ".join(default_scripts),)
                raise UserFeedback("No script file named '%s' found.\nAvailable built-in scripts: %s" % args)

        else:
            with open(scriptfile, "r") as f:
                js = f.read()

        args = {
                'width' : self.setting('width'),
                'height' : self.setting('height'),
                'port' : port,
                'name' : self.input_data.baserootname(),
                'ext' : self.setting('image-ext'),
                'cell_timeout' : self.setting('cell-timeout')
                }

        with open(scriptfile, "w") as f:
            f.write(js % args)

        with open(cellmetafile, "w") as f:
            json.dump(cellmetas, f)

    def launch_ipython(self, env):
        command = ['ipython', 'notebook', '--no-browser']
        command.extend(self.parse_additional_ipython_args())

        if self.setting('ipython-dir'):
            wd = self.setting('ipython-dir')
        else:
            wd = self.parent_work_dir()

        self.log_debug("About to run ipython command: '%s' in '%s'" % (' '.join(command), wd))
        proc = subprocess.Popen(command, shell=False,
                                    cwd=wd,
                                    stdin=None,
                                    stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE,
                                    env=env)

        self.log_debug("Reading from stderr of ipython command...")
        count = 0
        while True:
            count += 1
            if count > 100:
                raise Exception("IPython notebook failed to start.")

            line = proc.stderr.readline()
            self.log_debug(line)

            if "The IPython Notebook is running" in line:
                m = re.search("(localhost|[0-9\.]+):([0-9]{4})", line)
                port = m.groups()[1]

            if "Use Control-C to stop this server" in line:
                break

            if "ImportError" in line:
                raise Exception(line)

        # TODO if process is not running => throw exception
        return proc, port

    def parse_additional_ipython_args(self):
        raw_ipython_args = self.setting('ipython-args')
        if raw_ipython_args:
            if isinstance(raw_ipython_args, str):
                user_ipython_args = raw_ipython_args.split()
            elif isinstance(raw_ipython_args, list):
                assert isinstance(raw_ipython_args[0], str)
                user_ipython_args = raw_ipython_args
            else:
                raise UserFeedback("ipython-args must be a string or list of strings")
            return user_ipython_args
        else:
            return []

    def process(self):
        env = self.setup_env()

        wd = self.parent_work_dir()

        with open(self.input_data.storage.data_file(), "r") as f:
            nb = json.load(f)

        stdout = None
        output = {}
        output['cellmetas'] = []
        output['cellimages'] = []
        output['images-by-name'] = {}

        for cell in nb['worksheets'][0]['cells']:
            output['cellmetas'].append(cell['metadata'])

        ws = self.workspace()
        if os.path.exists(ws):
            self.log_debug("already have workspace '%s'" % os.path.abspath(ws))
        else:
            self.populate_workspace()

        # launch ipython notebook
        ipython_proc, port = self.launch_ipython(env)

        try:
            self.configure_casper_script(wd, port, output['cellmetas'])
    
            ## run casper script
            command = self.command_string()
            proc, stdout = self.run_command(command, self.setup_env())
            self.handle_subprocess_proc_return(command, proc.returncode, stdout)

        finally:
            # shut down ipython notebook
            os.kill(ipython_proc.pid, 9)


        if self.setting('add-new-files'):
            self.add_new_files()

        docname = self.doc.output_data().baserootname()
        i = 0
        for doc in sorted(self.doc.children):
            m = re.match("%s--([0-9]+)" % docname, doc.key)
            if m:
                assert i == int(m.groups()[0])
                output['cellimages'].append(doc.key)
                cellmeta_for_image = output['cellmetas'][i]
                if 'name' in cellmeta_for_image:
                    cellname = cellmeta_for_image['name']
                    output['images-by-name'][cellname] = doc.key
                i += 1

        if self.ext == ".txt":
            self.output_data.set_data(stdout)
        else:
            self.output_data.set_data(json.dumps(output))




from dexy.filters.process import SubprocessFilter
import dexy.exceptions

class OrgModeFilter(SubprocessFilter):
    """
    Convert .org files to other formats.
    """
    aliases = ['org']
    _settings = {
            'executable' : 'emacs',
            'output' : True,
            'input-extensions' : ['.org', '.txt'],
            'output-extensions' : ['.txt', '.html', '.tex', '.pdf', '.odt'],
            'command-string' : """%(prog)s --batch %(args)s --eval "(progn \\
(find-file \\"%(script_file)s\\") \\
(%(export_command)s 1) \\
(kill-buffer) \\
)"
"""
            }

    def command_string_args(self):
        if self.ext == '.txt':
            export_command = "org-export-as-ascii"
        elif self.ext == '.html':
            export_command = "org-export-as-html"
        elif self.ext == '.tex':
            export_command = "org-export-as-latex"
        elif self.ext == '.pdf':
            export_command = "org-export-as-pdf"
        elif self.ext == '.odt':
            export_command = "org-export-as-odt"
        else:
            msg = "unsupported extension %s"
            msgargs = (self.ext)
            raise dexy.exceptions.InternalDexyProblem(msg % msgargs)

        args = self.default_command_string_args()
        args['export_command'] = export_command
        return args



from dexy.data import Generic
try:
    from bs4 import BeautifulSoup
    AVAILABLE = True
except ImportError:
    AVAILABLE = False

class BeautifulSoupData(Generic):
    """
    Allow querying HTML using BeautifulSoup.
    """
    aliases = ['bs4']

    def soup(self):
        """
        Returns a BeautifulSoup object initialized with contents.
        """
        if not hasattr(self, '_soup'):
            self._soup = BeautifulSoup(self.data())
        return self._soup

    def select(self, query):
        """
        Returns a list of results from a CSS query.
        """
        return self.soup().select(query) 

    def select_one(self, query):
        """
        Returns a single result from a CSS query. Result must be unique.
        """
        selects = self.select(query)
        if not len(selects) == 1:
            raise Exception("Select on '%s' was not unique.")
        return selects[0]

    def __getitem__(self, key):
        return self.select_one(key)



from dexy.reporter import Reporter
from dexy.doc import Doc
from jinja2 import Environment
from jinja2 import FileSystemLoader
import operator
import os
import random
import shutil
import codecs

def link_to_doc(node):
    return """&nbsp;<a href="#%s">&darr; doc info</a>""" % node.output_data().websafe_key()

def link_to_doc_if_doc(node):
    if isinstance(node, Doc):
        return link_to_doc(node)
    else:
        return ""

class RunReporter(Reporter):
    """
    Returns info about a dexy run.
    """
    aliases = ['run']
    _settings = {
            'in-cache-dir' : True,
            'dir' : 'run',
            'filename' : 'index.html',
            'default' : True,
            "run-for-wrapper-states" : ["ran", "error"]
            }

    def run(self, wrapper):
        self.wrapper = wrapper
        self.remove_reports_dir(wrapper)
        self.copy_template_files()

        # If not too large, copy the log so it can be viewed in HTML
        self.wrapper.flush_logs()
        if os.path.getsize(self.wrapper.log_path()) < 500000:
            with codecs.open(self.wrapper.log_path(), 'r', encoding='UTF-8') as f:
                log_contents = f.read()
        else:
            log_contents = "Log file is too large to include in HTML. Look in %s" % self.wrapper.log_path()

        env_data = self.template_data()

        # add additional env elements - should these also be in plugins?
        env_data['wrapper'] = wrapper
        env_data['batch'] = wrapper.batch
        env_data['log_contents'] = log_contents

        def printable_args(args):
            return dict((k, v) for k, v in args.items() if not k in ('contents', 'wrapper'))

        def print_children(node, indent=0, extra=""):
            rand_id = random.randint(10000000,99999999)
            spaces = " " * 4 * indent
            nbspaces = "&nbsp;" * 4 * indent
            content = ""

            node_div = """%s<div data-toggle="collapse" data-target="#%s">%s%s%s%s</div>"""
            node_div_args = (spaces, rand_id, nbspaces, extra,
                    node.key_with_class(), link_to_doc_if_doc(node),)

            content += node_div % node_div_args
            content += """  %s<div id="%s" class="collapse">""" % (spaces, rand_id)

            for child in list(node.inputs):
                if not "Artifact" in child.__class__.__name__:
                    content += print_children(child, indent+1, "&rarr;")

            for child in node.children:
                if not "Artifact" in child.__class__.__name__:
                    content += print_children(child, indent+1)

            content += "  %s</div>" % spaces
            return content 

        env_data['print_children'] = print_children
        env_data['printable_args'] = printable_args

        env = Environment()
        env.loader = FileSystemLoader(os.path.dirname(__file__))
        template = env.get_template('template.html')

        template.stream(env_data).dump(self.report_file(), encoding="utf-8")
